"""
Representer base classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 18, 2011.
"""
from pyramid.compat import NativeIO
from pyramid.compat import bytes_
from pyramid.compat import text_

from everest.representers.utils import get_mapping_registry


__docformat__ = 'reStructuredText en'
__all__ = ['MappingResourceRepresenter',
           'RepresentationGenerator',
           'RepresentationParser',
           'Representer',
           'ResourceRepresenter',
           ]


class Representer(object):
    """
    Base class for all representers.

    A representer knows how to convert an object into a representation of a
    particular MIME type (content type) and vice versa.
    """
    #: The encoding to use to convert between bytes and string
    #: representations.
    encoding = 'utf-8'
    #: The registered MIME content type this representer is handling.
    content_type = None

    def from_string(self, string_representation, resource=None):
        """
        Extracts resource data from the given string and converts them to
        a new resource or updates the given resource from it.
        """
        stream = NativeIO(string_representation)
        return self.from_stream(stream, resource=resource)

    def from_bytes(self, bytes_representation, resource=None):
        """
        Extracts resource data from the given bytes representation and calls
        :method:`from_string` with the resulting text representation.
        """
        text = bytes_representation.decode(self.encoding)
        return self.from_string(text, resource=resource)

    def to_string(self, obj):
        """
        Converts the given resource to a string representation and returns
        it.
        """
        stream = NativeIO()
        self.to_stream(obj, stream)
        return text_(stream.getvalue(), encoding=self.encoding)

    def to_bytes(self, obj, encoding=None):
        """
        Converts the given resource to bytes representation in the encoding
        specified by :param:`encoding` and returns it.
        """
        if encoding is None:
            encoding = self.encoding
        text = self.to_string(obj)
        return bytes_(text, encoding=self.encoding)

    def from_stream(self, stream, resource=None):
        """
        Extracts resource data from the given stream and converts them to
        a new resource or updates the given resource from it.
        """
        raise NotImplementedError("Abstract method.")

    def to_stream(self, obj, stream):
        """
        Converts the given resource to a string representation and writes
        it to the given stream.
        """
        raise NotImplementedError("Abstract method.")


class ResourceRepresenter(Representer):
    """
    Abstract basee class for resource representers which know how to convert
    resource representations into resources and back.
    """
    def __init__(self, resource_class):
        Representer.__init__(self)
        self.resource_class = resource_class

    def from_stream(self, stream, resource=None):
        data_el = self.data_from_stream(stream)
        return self.resource_from_data(data_el, resource=resource)

    def to_stream(self, resource, stream):
        data_el = self.resource_to_data(resource)
        self.data_to_stream(data_el, stream)

    @classmethod
    def create_from_resource_class(cls, rc):
        """
        Factory method creating a new representer from the given registered
        resource.
        """
        raise NotImplementedError('Abstract method.')

    def data_from_stream(self, stream):
        """
        Extracts resource data from the given stream.
        """
        raise NotImplementedError('Abstract method.')

    def data_to_stream(self, data_element, stream):
        """
        Writes resource data to the given stream.
        """
        raise NotImplementedError('Abstract method.')

    def data_from_string(self, text):
        """
        Converts the given text representation to resource data.

        :returns: object implementing
            :class:`everest.representers.interfaces.IExplicitDataElement`
        """
        stream = NativeIO(text)
        return self.data_from_stream(stream)

    def data_from_bytes(self, byte_representation):
        """
        Converts the given bytes representation to resource data.
        """
        text = byte_representation.decode(self.encoding)
        return self.data_from_string(text)

    def data_to_string(self, data_element):
        """
        Converts the given data element into a string representation.

        :param data_element: object implementing
            :class:`everest.representers.interfaces.IExplicitDataElement`
        :returns: string representation (using the MIME content type
            configured for this representer)
        """
        stream = NativeIO()
        self.data_to_stream(data_element, stream)
        return stream.getvalue()

    def data_to_bytes(self, data_element, encoding=None):
        """
        Converts the given data element into a string representation using
        the :method:`data_to_string` method and encodes the resulting
        text with the given encoding.
        """
        if encoding is None:
            encoding = self.encoding
        text = self.data_to_string(data_element)
        return bytes_(text, encoding=encoding)

    def resource_from_data(self, data, resource=None):
        """
        Converts the given resource data to a resource.

        :param resource: If given, this resource will be updated from the
          given data; otherwise, a new resource will be created.
        :returns: object implementing
          :class:`everest.resources.interfaces.IResource`
        """
        raise NotImplementedError('Abstract method.')

    def resource_to_data(self, resource):
        """
        Converts the given resource to resource data.

        :returns: resource data object which can be passed to the *_from_data
            methods
        """
        raise NotImplementedError('Abstract method.')

    def configure(self, options=None):
        """
        Configures the options for this representer.
        """
        raise NotImplementedError('Abstract method.')


class MappingResourceRepresenter(ResourceRepresenter):
    """
    Base class for resource representers that use configurable attribute
    mappings to perform conversions from resource representations and back.

    The conversion is performed using four independent and highly customizable
    helper objects:

     1. The *representation parser* responsible for converting the
        representation into a data element tree;
     2. The *data element parser* responsible for converting a data element
        tree into a resource;
     3. The *data element generator* responsible for converting a resource
        into a data element tree; and
     4. the *representation generator* responsible for converting the data
        element tree into a representation.
    """
    def __init__(self, resource_class, mapping):
        ResourceRepresenter.__init__(self, resource_class)
        self._mapping = mapping

    @classmethod
    def create_from_resource_class(cls, resource_class):
        """
        Creates a new representer for the given resource class.

        The representer obtains a reference to the (freshly created or looked
        up) mapping for the resource class.
        """
        mp_reg = get_mapping_registry(cls.content_type)
        mp = mp_reg.find_or_create_mapping(resource_class)
        return cls(resource_class, mp)

    @classmethod
    def make_mapping_registry(cls):
        raise NotImplementedError('Abstract method.')

    def data_from_stream(self, stream):
        """
        Creates a data element reading a representation from the given stream.

        :returns: object implementing
            :class:`everest.representers.interfaces.IExplicitDataElement`
        """
        parser = self._make_representation_parser(stream, self.resource_class,
                                                  self._mapping)
        return parser.run()

    def data_to_stream(self, data_element, stream):
        """
        Writes the given data element to the given stream.
        """
        generator = \
            self._make_representation_generator(stream, self.resource_class,
                                                self._mapping)
        generator.run(data_element)

    def resource_from_data(self, data_element, resource=None):
        """
        Converts the given data element to a resource.

        :param data_element: object implementing
            :class:`everest.representers.interfaces.IExplicitDataElement`
        """
        return self._mapping.map_to_resource(data_element, resource=resource)

    def resource_to_data(self, resource):
        """
        Extracts managed attributes from a resource and constructs a data
        element for serialization from it.

        Default representer behavior:
         * Top-level member and collections resource attributes are
           represented as links.
         * Nested member resource attributes are represented as links,
           nested collection resource attributes are ignored (building a link
           may require iterating over the collection).

        The default behavior can be configured with the "representer"
        directive (:func:`everest.configuration.representer`) by means of the
        "write_as_link" and "ignore" options of representer configuration
        objects (:class:`everest.representers.config.RepresenterConfiguration`).
        """
        return self._mapping.map_to_data_element(resource)

    def configure(self, options=None, attribute_options=None): # pylint: disable=W0221
        """
        Configures the options and attribute options of the mapping associated
        with this representer with the given dictionaries.

        :param dict options: configuration options for the mapping associated
          with this representer.
        :param dict attribute_options: attribute options for the mapping
          associated with this representer.
        """
        self._mapping.update(options=options,
                             attribute_options=attribute_options)
#        self._mapping = \
#                self._mapping.clone(options=options,
#                                    attribute_options=attribute_options)

    def with_updated_configuration(self, options=None,
                                   attribute_options=None):
        """
        Returns a context in which this representer is updated with the
        given options and attribute options.
        """
        return self._mapping.with_updated_configuration(options=options,
                                                        attribute_options=
                                                            attribute_options)

    def _make_representation_parser(self, stream, resource_class, mapping):
        """
        Creates a representation parser from the given arguments. This parser
        has a `run` method which reads and parses the serialized resource from
        the given stream.
        """
        raise NotImplementedError('Abstract method.')

    def _make_representation_generator(self, stream, resource_class, mapping):
        """
        Creates a representation generator from the given arguments. This
        generator has a `run` method which writes out the serialized
        resource representation to the given stream.
        """
        raise NotImplementedError('Abstract method.')


class _RepresentationHandler(object):
    """
    Base class for classes handling a representation stream.
    """
    def __init__(self, stream, resource_class, mapping):
        """
        :param stream: stream object for the representation data.
        :param resource_class: registered member or collection resource
          class.
        :param mapping: attribute mapping for the given resource class,
          instance of :class:`everest.representers.mapping.Mapping`.
        """
        self._stream = stream
        self._resource_class = resource_class
        self._mapping = mapping
        self.__config = {}

    def get_option(self, option, default=None):
        """
        Returns the specified representer configuration option or the given
        default, if the option was not configured.
        """
        return self.__config.get(option, default)

    def set_option(self, option, value):
        """
        Sets the specified representer configuration option to the given
        value.
        """
        self.__config[option] = value


class RepresentationParser(_RepresentationHandler):
    """
    Abstract base class for classes that parse representations.
    """
    def run(self):
        """
        :return: The data element tree parsed from the handled stream.
        """
        raise NotImplementedError('Abstract method.')


class RepresentationGenerator(_RepresentationHandler):
    """
    Abstract base class for classes that generate representations.
    """
    def run(self, data_element):
        """
        :param data_element: The data element tree to be serialized.
        """
        raise NotImplementedError('Abstract method.')



"""
Entity state management.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Mar 14, 2013.
"""
from everest.entities.attributes import get_domain_class_attribute_iterator
from everest.entities.attributes import get_domain_class_attribute_names
from everest.utils import get_nested_attribute
from everest.utils import set_nested_attribute
from pyramid.compat import iteritems_
from weakref import ref

__docformat__ = 'reStructuredText en'
__all__ = ['ENTITY_STATUS',
           'EntityState',
           ]


class ENTITY_STATUS(object):
    """
    Entity status flags.
    """
    CLEAN = 'CLEAN'
    NEW = 'NEW'
    DELETED = 'DELETED'
    DIRTY = 'DIRTY'


class EntityState(object):
    """
    Tracks entity status, persistency, and state data.

    Initially, an object is marked as NEW (freshly instantiated) or CLEAN
    (freshly fetched from repository).

    Only a weak reference to the tracked object is stored to avoid circular
    references.

    Not all status transitions are allowed.
    """
    # FIXME: Need a proper state diagram here or drop tracking alltogether.
    __allowed_transitions = set([(None, ENTITY_STATUS.NEW),
                                 (None, ENTITY_STATUS.CLEAN),
                                 (None, ENTITY_STATUS.DELETED),
                                 (ENTITY_STATUS.NEW, ENTITY_STATUS.CLEAN),
                                 (ENTITY_STATUS.NEW, ENTITY_STATUS.DELETED),
                                 (ENTITY_STATUS.DELETED, ENTITY_STATUS.CLEAN),
                                 (ENTITY_STATUS.DELETED, ENTITY_STATUS.NEW),
                                 (ENTITY_STATUS.CLEAN, ENTITY_STATUS.DIRTY),
                                 (ENTITY_STATUS.CLEAN, ENTITY_STATUS.DELETED),
                                 (ENTITY_STATUS.CLEAN, ENTITY_STATUS.NEW),
                                 (ENTITY_STATUS.DIRTY, ENTITY_STATUS.CLEAN),
                                 (ENTITY_STATUS.DIRTY, ENTITY_STATUS.DELETED),
                                 ])

    def __init__(self, entity, unit_of_work):
        self.__entity_ref = ref(entity)
        self.__uow_ref = ref(unit_of_work)
        self.__status = None
        self.__clean_data = self.data
        #: Flag indicating if this state has been flushed to the backend.
        self.is_persisted = False

    @classmethod
    def manage(cls, entity, unit_of_work):
        """
        Manages the given entity under the given Unit Of Work.

        If `entity` is already managed by the given Unit Of Work, nothing
        is done.

        :raises ValueError: If the given entity is already under management
          by a different Unit Of Work.
        """
        if hasattr(entity, '__everest__'):
            if not unit_of_work is entity.__everest__.unit_of_work:
                raise ValueError('Trying to register an entity that has been '
                                 'registered with another session!')
        else:
            entity.__everest__ = cls(entity, unit_of_work)

    @classmethod
    def release(cls, entity, unit_of_work):
        """
        Releases the given entity from management under the given Unit Of
        Work.

        :raises ValueError: If `entity` is not managed at all or is not
          managed by the given Unit Of Work.
        """
        if not hasattr(entity, '__everest__'):
            raise ValueError('Trying to unregister an entity that has not '
                             'been registered yet!')
        elif not unit_of_work is entity.__everest__.unit_of_work:
            raise ValueError('Trying to unregister an entity that has been '
                             'registered with another session!')
        delattr(entity, '__everest__')

    @classmethod
    def get_state(cls, entity):
        try:
            return entity.__everest__
        except AttributeError:
            raise ValueError('Trying to obtain state for un-managed entity.')

    @classmethod
    def get_state_data(cls, entity):
        """
        Returns the state data for the given entity.

        This also works for unmanaged entities.
        """
        attrs = get_domain_class_attribute_iterator(type(entity))
        return dict([(attr,
                      get_nested_attribute(entity, attr.entity_attr))
                     for attr in attrs
                     if not attr.entity_attr is None])

    @classmethod
    def set_state_data(cls, entity, data):
        """
        Sets the state data for the given entity to the given data.

        This also works for unmanaged entities.
        """
        attr_names = get_domain_class_attribute_names(type(entity))
        nested_items = []
        for attr, new_attr_value in iteritems_(data):
            if not attr.entity_attr in attr_names:
                raise ValueError('Can not set attribute "%s" for entity '
                                 '"%s".' % (attr.entity_attr, entity))
            if '.' in attr.entity_attr:
                nested_items.append((attr, new_attr_value))
                continue
            else:
                setattr(entity, attr.entity_attr, new_attr_value)
        for attr, new_attr_value in nested_items:
            try:
                set_nested_attribute(entity, attr.entity_attr, new_attr_value)
            except AttributeError as exc:
                if not new_attr_value is None:
                    raise exc

    @classmethod
    def transfer_state_data(cls, source_entity, target_entity):
        """
        Transfers instance state data from the given source entity to the
        given target entity.
        """
        state_data = cls.get_state_data(source_entity)
        cls.set_state_data(target_entity, state_data)

    def __get_data(self):
        """
        Returns state data for the given entity of the given class.

        :param entity: Entity to obtain the state data from.
        :returns: Dictionary mapping attributes to attribute values.
        """
        ent = self.__entity_ref()
        return self.get_state_data(ent)

    def __set_data(self, data):
        """
        Sets the given state data on the given entity of the given class.

        :param data: State data to set.
        :type data: Dictionary mapping attributes to attribute values.
        :param entity: Entity to receive the state data.
        """
        ent = self.__entity_ref()
        self.set_state_data(ent, data)

    data = property(__get_data, __set_data)

    @property
    def clean_data(self):
        return self.__clean_data

    def __get_status(self):
        status = self.__status
        if status == ENTITY_STATUS.CLEAN:
            if self.data != self.__clean_data:
                status = ENTITY_STATUS.DIRTY
        return status

    def __set_status(self, status):
        if not (self.__get_status(), status) in self.__allowed_transitions:
            raise ValueError('Invalid status transition %s -> %s.'
                             % (self.__status, status))
        self.__status = status
        if status == ENTITY_STATUS.CLEAN:
            self.__clean_data = self.data

    #: The current status. One of the `ENTITY_STATUS` constants.
    status = property(__get_status, __set_status)

    @property
    def unit_of_work(self):
        return self.__uow_ref()

    @property
    def entity(self):
        return self.__entity_ref()




"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2012.
"""
from everest.entities.utils import get_root_aggregate
from everest.exceptions import MultipleResultsException
from everest.exceptions import NoResultsException
from everest.querying.base import CqlExpressionList
from everest.querying.filtering import CqlFilterExpression
from everest.querying.operators import ASCENDING
from everest.querying.operators import DESCENDING
from everest.querying.operators import GREATER_OR_EQUALS
from everest.querying.operators import GREATER_THAN
from everest.querying.operators import LESS_OR_EQUALS
from everest.querying.operators import LESS_THAN
from everest.querying.specifications import AscendingOrderSpecification
from everest.querying.specifications import DescendingOrderSpecification
from everest.querying.specifications import ValueEqualToFilterSpecification
from everest.repositories.memory.querying import EvalFilterExpression
from everest.repositories.memory.querying import EvalOrderExpression
from everest.repositories.rdb.testing import RdbTestCaseMixin
from everest.resources.staging import StagingAggregate
from everest.testing import EntityTestCase
from everest.testing import Pep8CompliantTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.testing import create_entity
from operator import and_ as operator_and
from operator import or_ as operator_or

__docformat__ = 'reStructuredText en'
__all__ = ['CqlExpressionTestCase',
           'EvalExpressionTestCase',
           'MemoryQueryTestCase',
           'MemoryRepositoryQueryTestCase',
           'RdbSessionQueryTestCase',
           ]


class CqlExpressionTestCase(Pep8CompliantTestCase):
    def test_cql_expression_combination(self):
        cql_expr = CqlFilterExpression('foo', LESS_THAN.name, '1')
        expr_str = 'foo:less-than:1'
        self.assert_equal(str(cql_expr), expr_str)
        cql_exprs = CqlExpressionList([cql_expr])
        and_expr_str = '~'.join((expr_str, expr_str))
        self.assert_equal(str(operator_and(cql_expr, cql_expr)),
                          and_expr_str)
        self.assert_equal(str(operator_and(cql_expr, cql_exprs)),
                          and_expr_str)
        self.assert_raises(TypeError, operator_and, cql_expr, None)
        self.assert_equal(str(operator_and(cql_exprs, cql_expr)),
                          and_expr_str)
        self.assert_equal(str(operator_and(cql_exprs, cql_exprs)),
                          and_expr_str)
        self.assert_raises(TypeError, operator_and, cql_exprs, None)
        cql_or_expr = operator_or(cql_expr, cql_expr)
        self.assert_equal(str(cql_or_expr), "%s,1" % expr_str)
        self.assert_raises(ValueError, operator_or, cql_expr,
                           CqlFilterExpression('bar', GREATER_THAN.name, '1'))

    def test_cql_expression_negation(self):
        inv_expr = ~CqlFilterExpression('foo', LESS_THAN.name, '1')
        self.assert_equal(inv_expr.op_name, GREATER_OR_EQUALS.name)
        inv_inv_expr = ~inv_expr
        self.assert_equal(inv_inv_expr.op_name, LESS_THAN.name)
        inv_expr = ~CqlFilterExpression('foo', GREATER_THAN.name, '1')
        self.assert_equal(inv_expr.op_name, LESS_OR_EQUALS.name)
        inv_inv_expr = ~inv_expr
        self.assert_equal(inv_inv_expr.op_name, GREATER_THAN.name)
        invalid_expr = CqlFilterExpression('foo', 'bar', '1')
        self.assert_raises(ValueError, invalid_expr.__invert__)


class EvalExpressionTestCase(Pep8CompliantTestCase):
    def test_filter_expr(self):
        expr0 = EvalFilterExpression(ValueEqualToFilterSpecification('id', 0))
        expr1 = \
         EvalFilterExpression(ValueEqualToFilterSpecification('text', 'text'))
        and_expr = expr0 & expr1
        or_expr = expr0 | expr1
        not_expr = ~expr0
        ent0 = MyEntity(id=0, text='text')
        ent1 = MyEntity(id=1, text='text')
        ents = [ent0, ent1]
        self.assert_equal(list(and_expr(ents)), [ent0])
        self.assert_equal(set(or_expr(ents)), set([ent0, ent1]))
        self.assert_equal(list(not_expr(ents)), [ent1])


class _BaseQueryTestCase(EntityTestCase):
    package_name = 'everest.tests.complete_app'

    def set_up(self):
        EntityTestCase.set_up(self)
        self._ent0 = create_entity(entity_id=0, entity_text='foo0')
        self._ent1 = create_entity(entity_id=1, entity_text='foo1')
        self._aggregate = get_root_aggregate(IMyEntity)
        self._aggregate.add(self._ent0)
        self._aggregate.add(self._ent1)

    @property
    def _query(self):
        return self._aggregate.query()

    def test_basics(self):
        self.assert_equal(set(self._query.all()),
                          set([self._ent0, self._ent1]))
        self.assert_equal(len(list(self._query)), 2)
        self.assert_equal(self._query.count(), 2)
        self.assert_equal(len(list(self._query.slice(0, 1))), 1)
        self.assert_equal(self._query.slice(0, 1).count(), 2)

    def test_one(self):
        self.assert_true(self._query.filter_by(id=1).one() is self._ent1)
        self.assert_raises(MultipleResultsException, self._query.one)
        self.assert_raises(NoResultsException,
                           self._query.filter_by(id=-1).one)

    def test_filter(self):
        q = self._query.filter_by(id=1).filter_by(text='foo1')
        self.assert_true(q.one() is self._ent1)

    def test_order_by(self):
        ent2 = create_entity(entity_id=2, entity_text='foo3')
        ent3 = create_entity(entity_id=3, entity_text='foo3')
        self._aggregate.add(ent2)
        self._aggregate.add(ent3)
        q = self._query.order_by(('text', ASCENDING), ('id', DESCENDING))
        last_ent = q.all()[-1]
        self.assert_true(last_ent is ent2)
        self.assert_raises(ValueError, self._query.order_by, ('text', None))

    def _test_order(self, txt_expr, id_expr):
        ent2 = create_entity(entity_id=2, entity_text='foo3')
        ent3 = create_entity(entity_id=3, entity_text='foo3')
        self._aggregate.add(ent2)
        self._aggregate.add(ent3)
        q = self._query.order(txt_expr)
        self.assert_true(next(iter(q)) is self._ent0)
        q = q.order(None)
        q = q.order(txt_expr).order(id_expr)
        last_ent = q.all()[-1]
        self.assert_true(last_ent is ent2)


class MemoryRepositoryQueryTestCase(_BaseQueryTestCase):
    config_file_name = 'configure_no_rdb.zcml'

    def test_order(self):
        txt_spec = AscendingOrderSpecification('text')
        txt_expr = EvalOrderExpression(txt_spec)
        id_spec = DescendingOrderSpecification('id')
        id_expr = EvalOrderExpression(id_spec)
        self._test_order(txt_expr, id_expr)


class RdbSessionQueryTestCase(RdbTestCaseMixin, _BaseQueryTestCase):
    def test_order(self):
        txt_expr = MyEntity.text.asc()
        id_expr = MyEntity.id.desc()
        self._test_order(txt_expr, id_expr)


class MemoryQueryTestCase(EntityTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_basics(self):
        agg = StagingAggregate(MyEntity)
        ent0 = MyEntity(id=0, text='text0')
        ent1 = MyEntity(id=1, text='text1')
        agg.add(ent0)
        agg.add(ent1)
        q = agg.query()
        filter_expr = \
            EvalFilterExpression(ValueEqualToFilterSpecification('id', 0))
        self.assert_equal(q.filter(filter_expr).all(), [ent0])
        self.assert_equal(len(q.slice(1, 2).all()), 1)
        self.assert_equal(q.slice(1, 2).count(), 2)
        order_expr = EvalOrderExpression(AscendingOrderSpecification('text'))
        q = q.order(order_expr)
        self.assert_equal(q.all()[0].text, 'text0')



"""
Service.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 27, 2011.
"""
from everest.repositories.utils import as_repository
from everest.resources.base import Resource
from everest.resources.interfaces import IService
from pyramid.compat import iterkeys_
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['Service',
           ]


@implementer(IService)
class Service(Resource):
    """
    The service resource class.

    The service resource is placed at the root of the resource tree and
    provides traversal (=URL) access to all exposed collection resources.
    """
    relation = 'service'

    def __init__(self):
        Resource.__init__(self)
        # Setting the __name__ to None ensures a leading slash in model paths.
        self.__name__ = None
        # Collects all interfaces managed by the service.
        self.__registered_interfaces = set()
        # Maps collection names to resource interfaces.
        self.__collections = {}
        self.__started = False

    def register(self, irc):
        """
        Registers the given resource interface with this service.
        """
        if self.__started:
            raise RuntimeError("Can not register new resource interfaces when "
                               "the service has been started.")
        self.__registered_interfaces.add(irc)

    def start(self):
        """
        Starts the service.

        This adds all registered resource interfaces to the service. Multiple
        calls to this method will only perform the startup once.
        """
        if not self.__started:
            self.__started = True
            for irc in self.__registered_interfaces:
                self.add(irc)

    def stop(self):
        """
        Stops the service.
        """
        if self.__started:
            self.__started = False
            self.__collections.clear()

    def __getitem__(self, key):
        """
        Overrides __getitem__ to return a clone of the requested collection.

        :param key: collection name.
        :type key: str
        :returns: object implementing
          :class:`everest.resources.interfaces.ICollectionResource`.
        """
        irc = self.__collections[key]
        repo = as_repository(irc)
        coll = repo.get_collection(irc)
        coll.__parent__ = self
        return coll

    def __len__(self):
        return len(self.__collections)

    def __iter__(self):
        for key in iterkeys_(self.__collections):
            yield self.__getitem__(key)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           'started' if self.__started else 'not started')

    def add(self, irc):
        repo = as_repository(irc)
        coll = repo.get_collection(irc)
        if coll.__name__ in self.__collections:
            raise ValueError('Root collection for collection name %s '
                             ' already exists.' % coll.__name__)
        # We need to tell the repository that the service object should be
        # set as the parent for the newly created collection.
        repo.set_collection_parent(irc, self)
        # Update the collection name -> registered resource mapping.
        self.__collections[coll.__name__] = irc

    def remove(self, irc):
        repo = as_repository(irc)
        coll = repo.get_collection(irc)
        del self.__collections[coll.__name__]

    def get(self, name, default=None):
        try:
            coll = self.__getitem__(name)
        except KeyError:
            coll = default
        return coll



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 18, 2012.
"""
from pyramid.testing import DummyRequest
import pytest

from everest.configuration import Configurator
from everest.interfaces import IResourceUrlConverter
from everest.mime import CsvMime
from everest.querying.base import EXPRESSION_KINDS
from everest.querying.interfaces import IFilterSpecificationFactory
from everest.querying.interfaces import IFilterSpecificationVisitor
from everest.querying.interfaces import IOrderSpecificationFactory
from everest.querying.interfaces import IOrderSpecificationVisitor
from everest.repositories.constants import REPOSITORY_TYPES
from everest.repositories.interfaces import IRepositoryManager
from everest.repositories.memory import Aggregate
from everest.representers.csv import CsvResourceRepresenter
from everest.representers.interfaces import IRepresenterRegistry
from everest.resources.interfaces import IService
from everest.resources.utils import get_collection_class
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.interfaces import IFoo
from everest.tests.simple_app.resources import FooMember


__docformat__ = 'reStructuredText en'
__all__ = ['TestConfigurator',
           ]


class Dummy(object):
    pass


NotAMember = Dummy
NotACollection = Dummy
NotAnInterface = Dummy
NotAnEntity = Dummy


class UnrelatedMember(FooMember):
    relation = None


class MyRepresenterClass(CsvResourceRepresenter):
    pass


class TestConfigurator(object):
    def test_registry_setup(self, simple_config):
        reg = simple_config.registry
        assert not reg.queryUtility(IRepositoryManager) is None
        assert not reg.queryUtility(IFilterSpecificationFactory) is None
        assert not reg.queryUtility(IOrderSpecificationFactory) is None
        assert not reg.queryUtility(IService) is None
        assert not reg.queryUtility(IFilterSpecificationVisitor,
                                    name=EXPRESSION_KINDS.CQL) is None
        assert not reg.queryUtility(IFilterSpecificationVisitor,
                                    name=EXPRESSION_KINDS.SQL) is None
        assert not reg.queryUtility(IFilterSpecificationVisitor,
                                    name=EXPRESSION_KINDS.EVAL) is None
        assert not reg.queryUtility(IOrderSpecificationVisitor,
                                    name=EXPRESSION_KINDS.CQL) is None
        assert not reg.queryUtility(IOrderSpecificationVisitor,
                                    name=EXPRESSION_KINDS.SQL) is None
        assert not reg.queryUtility(IOrderSpecificationVisitor,
                                    name=EXPRESSION_KINDS.EVAL) is None
        req = DummyRequest()
        assert not reg.queryAdapter(req, IResourceUrlConverter) is None

    @pytest.mark.parametrize('args,options',
                             [((NotAnInterface, FooMember, FooEntity),
                               dict(expose=False)),
                              ((IFoo, NotAMember, FooEntity),
                               dict(expose=False)),
                              ((IFoo, FooMember, NotAnEntity),
                               dict(expose=False)),
                              ((IFoo, FooMember, FooEntity),
                               dict(expose=False, collection=NotACollection)),
                              ((IFoo, UnrelatedMember, FooEntity),
                               dict(expose=False)),
                              ((IFoo, FooMember, FooEntity),
                               dict(expose=False, repository='UNKNOWN')),
                              ((IFoo, FooMember, FooEntity),
                               dict(expose=True)),
                              ]
                             )
    def test_add_resource(self, simple_config, args, options):
        with pytest.raises(ValueError):
            simple_config.add_resource(*args, **options)

    @pytest.mark.parametrize('title', ['myfoos'])
    def test_add_resource_with_collection_title(self, simple_config, title):
        simple_config.add_resource(IFoo, FooMember, FooEntity, expose=False,
                                   collection_title=title)
        assert get_collection_class(IFoo).title == title

    @pytest.mark.parametrize('root_name', ['myfoos'])
    def test_add_resource_with_root_name(self, simple_config, root_name):
        simple_config.add_resource(IFoo, FooMember, FooEntity, expose=True,
                                   collection_root_name=root_name)
        assert get_collection_class(IFoo).root_name == root_name

    def test_add_resource_with_rdb_repo(self, simple_config):
        simple_config.add_resource(IFoo, FooMember, FooEntity, expose=False,
                                   repository=REPOSITORY_TYPES.RDB)
        reg = simple_config.registry
        repo_mgr = reg.queryUtility(IRepositoryManager)
        assert not repo_mgr.get(REPOSITORY_TYPES.RDB) is None

    def test_have_memory_repo(self, simple_config):
        reg = simple_config.registry
        repo_mgr = reg.queryUtility(IRepositoryManager)
        assert not repo_mgr.get(REPOSITORY_TYPES.MEMORY) is None

    def test_add_resource_with_filesystem_repo(self, simple_config):
        simple_config.add_resource(IFoo, FooMember, FooEntity, expose=False,
                                  repository=REPOSITORY_TYPES.FILE_SYSTEM)
        reg = simple_config.registry
        repo_mgr = reg.queryUtility(IRepositoryManager)
        assert not repo_mgr.get(REPOSITORY_TYPES.FILE_SYSTEM) is None

    def test_add_representer(self, simple_config):
        with pytest.raises(ValueError):
            simple_config.add_representer()
        with pytest.raises(ValueError):
            simple_config.add_representer(
                                    content_type=CsvMime,
                                    representer_class=CsvResourceRepresenter)

    def test_add_representer_with_representer_class(self, simple_config):
        simple_config.add_representer(representer_class=MyRepresenterClass)
        reg = simple_config.registry
        rpr_reg = reg.queryUtility(IRepresenterRegistry)
        assert rpr_reg.is_registered_representer_class(MyRepresenterClass)

    def test_add_resource_representer(self, simple_config):
        with pytest.raises(ValueError):
            simple_config.add_resource_representer(NotAMember, CsvMime)

    def test_custom_repository(self, simple_config):
        class MyMemoryAggregate(Aggregate):
            pass
        reg = simple_config.registry
        config = Configurator(registry=reg)
        config.add_memory_repository('test',
                                     aggregate_class=MyMemoryAggregate)
        repo_mgr = config.get_registered_utility(IRepositoryManager)
        repo = repo_mgr.get('test')
        config.add_resource(IFoo, FooMember, FooEntity,
                            collection_root_name="foos",
                            repository='test')
        with pytest.raises(RuntimeError):
            repo.get_collection(IFoo)
        with pytest.raises(RuntimeError):
            repo.get_aggregate(IFoo)
        repo.initialize()
        coll = repo.get_collection(IFoo)
        agg = coll.get_aggregate()
        assert isinstance(agg, MyMemoryAggregate)
        entity = FooEntity(id=1)
        agg.add(entity)
        assert agg.count() == 1
        assert list(agg.iterator())[0].id == entity.id
        assert agg.get_by_id(1).id == entity.id
        assert agg.get_by_slug('1').slug == entity.slug
        agg.remove(entity)
        assert agg.count() == 0



"""
Interfaces for everest.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 14, 2011.
"""
from zope.interface import Attribute # pylint: disable=E0611,F0401
from zope.interface import Interface # pylint: disable=E0611,F0401
from zope.schema import Bool # pylint: disable=E0611,F0401
from zope.schema import List # pylint: disable=E0611,F0401
from zope.schema import Text # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['IAtomEntryMime',
           'IAtomFeedMime',
           'IAtomMime',
           'IAtomRequest',
           'IAtomServiceMime',
           'ICsvMime',
           'IDataTraversalProxyAdapter',
           'IDataTraversalProxyAdapter',
           'ICsvRequest',
           'IHtmlRequest',
           'IHtmlMime',
           'IJsonMime',
           'IJsonRequest',
           'IUserMessage',
           'IUserMessageChecker',
           'IUserMessageNotifier',
           'IXmlMime',
           'IXmlRequest',
           'IXlsMime',
           'IXlsRequest',
           'IZipMime'
           ]

# no self, no __init__, no args  pylint: disable=E0213,W0232,E0211

class IJsonRequest(Interface):
    """Marker interface for a JSON request."""


class IAtomRequest(Interface):
    """Marker interface for an ATOM request."""


class IXmlRequest(Interface):
    """Marker interface for an XML request."""


class ICsvRequest(Interface):
    """Marker interface for an request."""


class IHtmlRequest(Interface):
    """Marker interface for a HTML request."""


class IXlsRequest(Interface):
    """Marker interface for an Excel request."""


class IZipRequest(Interface):
    """Marker interface for an Zip compressed request."""


class IMime(Interface):
    """Base Interface for all MIME content types."""
    mime_type_string = Text(title=u'MIME string for this MIME content type. '
                                   '(e.g., "application/json".')
    representer_name = Text(title=u'Name of the representer associated with '
                                   'this MIME content type (e.g., "json").')
    file_extensions = List(title=u'Known file extensions for this MIME '
                                  'content type.')

class IJsonMime(IMime):
    """Marker interface for JSON mime type."""


class IAtomMime(IMime):
    """Marker interface for ATOM mime type."""


class IXmlMime(IMime):
    """Marker interface for XML mime type."""


class ICsvMime(IMime):
    """Marker interface for CSV mime type."""


class IXlsMime(IMime):
    """Marker interface for Excel mime type."""


class IZipMime(IMime):
    """Marker interface for Zip compressed mime type."""


class IHtmlMime(IMime):
    """Marker interface for HTML mime type."""


class ITextPlainMime(IMime):
    """Marker interface for Plain Text mime type."""


class IAtomFeedMime(Interface):
    """Marker interface for ATOM feed mime type."""


class IAtomEntryMime(Interface):
    """Marker interface for ATOM entry mime type."""


class IAtomServiceMime(Interface):
    """Marker interface for ATOM service mime type."""


class IUserMessage(Interface):
    """
    Interface for the user message system resource.
    """
    text = Text(title=u'message text.')


class IUserMessageNotifier(Interface):
    """
    Interface for user message notifiers.
    """
    def notify(message_text):
        """
        Notifies all subscribers to
        :class:`everest.messaging.IUserMessage` of the given message
        and returns their collective vote.

        :param str message_text: message to notify subcribers about. The
          notifier will need to convert this to a :class:`UserMessage`
          instance.
        """


class IUserMessageChecker(Interface):
    """
    Interface for user message checkers.
    """
    def __call__(message):
        """
        This is required so that instances of user message checkers can
        serve as an adapter to a user message.

        :param message: message to check
        :type message: :class:`everest.entities.system.UserMessage`.
        """

    def check():
        """
        Evaluates the message held by this checker. Returns `False`,
        `True`, or `None` to signal abortion, continuation conditional on
        other checkers approval, or unconditional continuation to the caller.
        """

    vote = Bool(title=u'Result of the voting from all subscribed checkers.')


class IResourceUrlConverter(Interface):
    """
    Interface for resource <-> URL converters.
    """
    def url_to_resource(url):
        """Performs URL -> resource conversion."""
    def resource_to_url(resource):
        """Performs URL -> resource conversion."""


class IDataTraversalProxyFactory(Interface):
    """
    Interface for data traversal proxy factories.
    """
    def make_source_proxy(data, options=None):
        """
        Creates a source data traversal proxy.
        """

    def make_target_proxy(data, accessor,
                          manage_back_references=True, options=None):
        """
        Creates a target data traversal proxy.
        """

    def make_proxy(self, data, accessor, relationship_direction,
                   options=None):
        """
        Creates a data traversal proxy.
        """


class IDataTraversalProxyAdapter(IDataTraversalProxyFactory):
    """
    Interface for data traversal proxy adapters.
    """
    proxy_class = Attribute('The data traversal proxy class for this '
                            'adapter.')


class IRelationship(Interface):
    specification = Attribute('Filter specification for the objects '
                              'defined by this relationship.')

    def add(related, direction=None, safe=False):
        """
        Adds the given related object to the relationship.

        The add operation is performed on both ends of the relationship if
        appropriate entity attribute declarations have been made.

        :param related: object to ADD.
        :param direction: One of the constants defined in
          :class:`everest.constants.RELATIONSHIP_DIRECTIONS`. Indicates if
          the attribute of the relator (FORWARD), of the relatee (REVERSE),
          or both (BIDIRECTIONAL) should be updated.
        :param safe: Flag indicating if the ADD operation should
          only proceed if the object to add is not already in the
          relationship.
        """

    def remove(related, direction=None, safe=False):
        """
        Removes the given related object from the relationship.

        The remove operation is performed on both ends of the relationship if
        appropriate entity attribute declarations have been made.

        :param related: object to REMOVE.
        :param direction: One of the constants defined in
          :class:`everest.constants.RELATIONSHIP_DIRECTIONS`. Indicates if
          the attribute of the relator (FORWARD), of the relatee (REVERSE),
          or both (BIDIRECTIONAL) should be updated.
        :param safe: Flag indicating if the REMOVE operation should succeed
          even when the object being removed is not in the relationship.
        :raises ValueError: If :param:`related` is not in the relationship
          and :param:`safe` is not set.
        """

# pylint: enable=E0213,W0232,E0211



"""
Session for the rdb backend.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 8, 2013.
"""
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session as SaSession

from everest.constants import RELATION_OPERATIONS
from everest.entities.interfaces import IEntity
from everest.entities.traversal import AruVisitor
from everest.repositories.base import AutocommittingSessionMixin
from everest.repositories.base import Session
from everest.repositories.base import SessionFactory
from everest.repositories.state import EntityState
from everest.traversal import SourceTargetDataTreeTraverser
from zope.sqlalchemy import ZopeTransactionExtension # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['RdbAutocommittingSession',
           'RdbSession',
           'RdbSessionFactory',
           'ScopedSessionMaker',
           ]


class RdbSession(SaSession, Session):
    """
    Special session class adapting the SQLAlchemy session for everest.
    """
    IS_MANAGING_BACKREFERENCES = False

    def __init__(self, *args, **options):
        self.__repository = options.pop('repository')
        SaSession.__init__(self, *args, **options)

    def get_by_id(self, entity_class, id_key):
        return self.query(entity_class).get(id_key)

    def get_by_slug(self, entity_class, slug):
        # We don't have an optimization for access by slug here; returning
        # `None` indicates that a query should be run.
        return None

    def add(self, entity_class, data): # different signature pylint: disable=W0222
        if not IEntity.providedBy(data): # pylint: disable=E1101
            self.__run_traversal(entity_class, data, None,
                                 RELATION_OPERATIONS.ADD)
        else:
            SaSession.add(self, data)

    def remove(self, entity_class, data):
        if not IEntity.providedBy(data): # pylint: disable=E1101
            self.__run_traversal(entity_class, None, data,
                                 RELATION_OPERATIONS.REMOVE)
        else:
            SaSession.delete(self, data)

    def update(self, entity_class, data, target=None):
        if not IEntity.providedBy(data): # pylint: disable=E1101
            upd_ent = self.__run_traversal(entity_class, data, target,
                                           RELATION_OPERATIONS.UPDATE)
        else:
            upd_ent = SaSession.merge(self, data)
        return upd_ent

    def query(self, *entities, **options):
        # When called by everest from an aggregate, we use the counting query
        # class that attempts to fetch the total result count and the first
        # result page in one call here.
        query_cls = options.pop('query_class', Query)
        return query_cls(entities, self, **options)

    def reset(self):
        self.rollback()
        self.expunge_all()

    def __run_traversal(self, entity_class, source_data, target_data, rel_op):
        agg = self.__repository.get_aggregate(entity_class)
        trv = SourceTargetDataTreeTraverser.make_traverser(
                                    source_data, target_data, rel_op,
                                    accessor=agg,
                                    manage_back_references=False)
        vst = AruVisitor(entity_class,
                         add_callback=self.__add,
                         remove_callback=self.__remove,
                         update_callback=self.__update,
                         pass_path_to_callbacks=True)
        trv.run(vst)
        return vst.root

    def __add(self, entity_class, entity, path): # pylint: disable=W0613
        if len(path) == 0:
            SaSession.add(self, entity)

    def __remove(self, entity_class, entity, path): # pylint: disable=W0613
        if len(path) == 0:
            SaSession.delete(self, entity)

    def __update(self, entity_class, source_data, target_entity, path): # pylint: disable=W0613
        EntityState.set_state_data(target_entity, source_data)


class RdbAutocommittingSession(AutocommittingSessionMixin, RdbSession):
    def __init__(self, **kw):
        kw['autocommit'] = True
        super(RdbAutocommittingSession, self).__init__(**kw)


#: The scoped session maker. Instantiate this to obtain a thread local
#: session instance.
ScopedSessionMaker = scoped_session(sessionmaker(class_=RdbSession))


class RdbSessionFactory(SessionFactory):
    """
    Factory for RDB repository sessions.
    """
    def __init__(self, repository, counting_query_class):
        SessionFactory.__init__(self, repository)
        if self._repository.autocommit:
            # Use an autocommitting Session class with our session factory.
            self.__fac = scoped_session(
                                sessionmaker(class_=RdbAutocommittingSession))
        else:
            # Use the default Session factory.
            self.__fac = ScopedSessionMaker
        #: This is the (optimized, if the engine supports it) counting query
        #: class used for paged queries.
        self.counting_query_class = counting_query_class

    def configure(self, **kw):
        self.__fac.configure(**kw)

    def __call__(self, **kw):
        if not self.__fac.registry.has():
            self.__fac.configure(
                            autoflush=self._repository.autoflush,
                            repository=self._repository)
            if not self._repository.autocommit \
               and self._repository.join_transaction:
                # Enable the Zope transaction extension.
                self.__fac.configure(extension=ZopeTransactionExtension())
            else:
                # Disable extension otherwise.
                self.__fac.configure(extension=None)
        return self.__fac(**kw)



"""
Constants for the repositories package.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 25, 2013.
"""

__docformat__ = 'reStructuredText en'
__all__ = ['REPOSITORY_DOMAINS',
           'REPOSITORY_TYPES',
           ]


class REPOSITORY_TYPES(object):
    MEMORY = 'MEMORY'
    RDB = 'RDB'
    FILE_SYSTEM = 'FILE_SYSTEM'
    NO_SQL = 'NO_SQL'


class REPOSITORY_DOMAINS(object):
    ROOT = 'ROOT'
    SYSTEM = 'SYSTEM'



"""
Resource reference expression parser.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 31, 2014.
"""
from pyparsing import CaselessKeyword
from pyparsing import Combine
from pyparsing import Group
from pyparsing import Literal
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import alphanums
from pyparsing import alphas
from pyparsing import delimitedList
from pyparsing import replaceWith

from everest.constants import ResourceReferenceRepresentationKinds
from everest.entities.utils import identifier_from_slug
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import WRITE_AS_LINK_OPTION


__docformat__ = 'reStructuredText en'
__all__ = ['parse_refs',
           ]

TILDE_PAT = '~'
URL_PAT = ResourceReferenceRepresentationKinds.URL
INLINE_PAT = ResourceReferenceRepresentationKinds.INLINE
OFF_PAT = ResourceReferenceRepresentationKinds.OFF

URL = CaselessKeyword("url").setParseAction(replaceWith(URL_PAT))
INLINE = CaselessKeyword("inline").setParseAction(replaceWith(INLINE_PAT))
OFF = CaselessKeyword("off").setParseAction(replaceWith(OFF_PAT))


class RefsConverter(object):
    """
    Converter for a sequence of resource references.
    """
    @classmethod
    def convert(cls, toks):
        config = {}
        for rf in toks:
            config.update(cls.__convert_ref(rf))
        return config

    @classmethod
    def __convert_ref(cls, rf):
        names = [identifier_from_slug(token)
                 for token in rf.attribute.split('.')]
        config = {}
        # First, make sure that all parent attributes are configured INLINE.
        for idx in range(len(names[:-1])):
            key = tuple(names[:idx + 1])
            config[key] = {IGNORE_OPTION:False,
                           WRITE_AS_LINK_OPTION:False}
        is_off = rf.option == OFF_PAT
        opts = {IGNORE_OPTION:is_off}
        if not is_off:
            opts[WRITE_AS_LINK_OPTION] = rf.option == URL_PAT
        config[tuple(names)] = opts
        return config


colon = Literal(':')
attribute = Word(alphas, alphanums + '-')
identifier = \
    Combine(attribute + ZeroOrMore('.' + attribute)).setName('identifier')
option = (URL | INLINE | OFF)
ref = Group(identifier('attribute')
                  + colon.suppress()
                  + option('option'))
refs = delimitedList(ref, delim=TILDE_PAT)
refs.setParseAction(RefsConverter.convert)


def parse_refs(refs_string):
    return refs.parseString(refs_string)[0]



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 14, 2012.
"""
from everest.resources.staging import create_staging_collection
from everest.resources.utils import get_registered_collection_resources
from everest.resources.utils import get_resource_class_for_relation
from everest.resources.utils import get_resource_url
from everest.resources.utils import provides_collection_resource
from everest.resources.utils import provides_member_resource
from everest.resources.utils import provides_resource
from everest.testing import ResourceTestCase
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.resources import MyEntityMember
from everest.tests.complete_app.testing import create_collection
from everest.tests.complete_app.testing import create_entity


__docformat__ = 'reStructuredText en'
__all__ = ['ResourcesUtilsTestCase',
           ]


class ResourcesUtilsTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_get_resource_url(self):
        coll = create_collection()
        url = get_resource_url(coll)
        self.assert_not_equal(url.find('/my-entities'), -1)

    def test_provides_resource(self):
        coll = create_collection()
        mb = next(iter(coll))
        self.assert_true(provides_member_resource(type(mb)))
        self.assert_true(provides_member_resource(mb))
        self.assert_false(provides_member_resource(coll))
        self.assert_true(provides_resource(type(mb)))
        self.assert_true(provides_resource(mb))
        self.assert_true(provides_collection_resource(type(coll)))
        self.assert_true(provides_collection_resource(coll))
        self.assert_false(provides_collection_resource(mb))
        self.assert_true(provides_resource(type(coll)))
        self.assert_true(provides_resource(coll))
        self.assert_false(provides_resource(mb.get_entity()))

    def test_get_registered_collection_resources(self):
        colls = get_registered_collection_resources()
        self.assert_equal(len(colls), 4)

    def test_get_resource_from_relation(self):
        rel = MyEntityMember.relation
        self.assert_equal(get_resource_class_for_relation(rel),
                          MyEntityMember)

    def test_new_stage_collection(self):
        ent = create_entity(entity_id=2, entity_text='too2')
        mb = MyEntityMember.create_from_entity(ent)
        nscoll = create_staging_collection(IMyEntity)
        self.assert_equal(len(nscoll), 0)
        nscoll.add(mb)
        self.assert_equal(len(nscoll), 1)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 21, 2012.
"""
import glob
import os
import shutil
import tempfile
import zipfile

from pyramid.compat import NativeIO
from pyramid.compat import itervalues_

from everest.compat import BytesIO
from everest.mime import CsvMime
from everest.repositories.rdb.testing import RdbTestCaseMixin
from everest.representers.config import IGNORE_OPTION
from everest.resources.staging import create_staging_collection
from everest.resources.storing import ConnectedResourcesSerializer
from everest.resources.storing import build_resource_dependency_graph
from everest.resources.storing import dump_resource
from everest.resources.storing import dump_resource_to_files
from everest.resources.storing import dump_resource_to_zipfile
from everest.resources.storing import find_connected_resources
from everest.resources.storing import get_collection_filename
from everest.resources.storing import get_collection_name
from everest.resources.storing import load_collection_from_file
from everest.resources.storing import load_collection_from_stream
from everest.resources.storing import load_collection_from_url
from everest.resources.storing import load_into_collection_from_url
from everest.resources.storing import load_into_collections_from_zipfile
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from everest.resources.utils import get_root_collection
from everest.testing import ResourceTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityGrandchild
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityGrandchild
from everest.tests.complete_app.interfaces import IMyEntityParent


__docformat__ = 'reStructuredText en'
__all__ = ['ConnectedResourcesTestCase',
           'FileResourceIoTestCase',
           'ResourceDependencyGraphTestCase',
           'ResourceGraphTestCase',
           'StreamResourceIoTestCase',
           'ZipResourceIoTestCaseNoRdb',
           'ZipResourceIoTestCaseRdb',
           ]


def _make_test_entity_member():
    parent = MyEntityParent()
    entity = MyEntity(parent=parent)
    if parent.child is None:
        parent.child = entity
    child = MyEntityChild()
    entity.children.append(child)
    if child.parent is None:
        child.parent = entity
    grandchild = MyEntityGrandchild()
    child.children.append(grandchild)
    if grandchild.parent is None:
        grandchild.parent = child
    coll = create_staging_collection(IMyEntity)
    mb = coll.create_member(entity)
    parent.id = 0
    entity.id = 0
    child.id = 0
    grandchild.id = 0
    return mb


class ResourceGraphTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'
    _interfaces = [IMyEntityParent, IMyEntity, IMyEntityChild,
                   IMyEntityGrandchild]


class ResourceDependencyGraphTestCase(ResourceGraphTestCase):

    def test_dependency_graph(self):
        grph = build_resource_dependency_graph(self._interfaces)
        self.assert_equal(len(grph.nodes()), 4)
        entity_mb_cls = get_member_class(IMyEntity)
        entity_parent_mb_cls = get_member_class(IMyEntityParent)
        entity_child_mb_cls = get_member_class(IMyEntityChild)
        entity_grandchild_mb_cls = get_member_class(IMyEntityGrandchild)
        # Entity Parent resource deps should be empty.
        self.assert_equal(grph.neighbors(entity_parent_mb_cls), [])
        # Entity Child has Grandchild.
        self.assert_equal(grph.neighbors(entity_child_mb_cls),
                          [entity_grandchild_mb_cls])
        # Entity Grandchild has Child, but backrefs are excluded by default.
        self.assert_equal(grph.neighbors(entity_grandchild_mb_cls),
                          [])
        # Entity has Parent and Child.
        self.assert_equal(set(grph.neighbors(entity_mb_cls)),
                          set([entity_parent_mb_cls, entity_child_mb_cls]))


class ConnectedResourcesTestCase(ResourceGraphTestCase):

    def test_find_connected_with_member(self):
        member = _make_test_entity_member()
        ent_map = find_connected_resources(member)
        for ents in itervalues_(ent_map):
            self.assert_equal(len(ents), 1)

    def test_find_connected_with_collection(self):
        member = _make_test_entity_member()
        ent_map = find_connected_resources(member.__parent__)
        for ents in itervalues_(ent_map):
            self.assert_equal(len(ents), 1)

    def test_find_connected_with_deps(self):
        member = _make_test_entity_member()
        dep_grph = \
            build_resource_dependency_graph(self._interfaces,
                                            include_backrefs=True)
        ent_map = find_connected_resources(member,
                                            dependency_graph=dep_grph)
        # Backrefs should not make a difference since we check for duplicates.
        for ents in itervalues_(ent_map):
            self.assert_equal(len(ents), 1)

    def test_convert_to_strings(self):
        member = _make_test_entity_member()
        srl = ConnectedResourcesSerializer(CsvMime)
        rpr_map = srl.to_strings(member)
        self.assert_equal(len(rpr_map), 4)


class _ResourceIoTestCaseBase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'

    def set_up(self):
        ResourceTestCase.set_up(self)
        # We need to switch off non-standard resource attributes manually.
        self.config.add_resource_representer(
                    IMyEntity, CsvMime,
                    attribute_options=
                            {('children',):{IGNORE_OPTION:True}
                             })


class _ZipResourceIoTestCaseBase(_ResourceIoTestCaseBase):

    def test_load_from_zipfile(self):
        member = _make_test_entity_member()
        strm = BytesIO()
        dump_resource_to_zipfile(member, strm)
        colls = [
                 get_root_collection(IMyEntityParent),
                 get_root_collection(IMyEntity),
                 get_root_collection(IMyEntityChild),
                 get_root_collection(IMyEntityGrandchild),
                 ]
        self.assert_equal(len(colls[0]), 0)
        self.assert_equal(len(colls[1]), 0)
        self.assert_equal(len(colls[2]), 0)
        self.assert_equal(len(colls[3]), 0)
        load_into_collections_from_zipfile(colls, strm)
        self.assert_equal(len(colls[0]), 1)
        self.assert_equal(len(colls[1]), 1)
        self.assert_equal(len(colls[2]), 1)
        self.assert_equal(len(colls[3]), 1)


class ZipResourceIoTestCaseNoRdb(_ZipResourceIoTestCaseBase):
    config_file_name = 'configure_no_rdb.zcml'

    def test_load_from_zipfile_invalid_extension(self):
        strm = BytesIO()
        zipf = zipfile.ZipFile(strm, 'w')
        coll_name = get_collection_name(get_collection_class(IMyEntity))
        zipf.writestr('%s.foo' % coll_name, '')
        zipf.close()
        colls = [get_root_collection(IMyEntity)]
        with self.assert_raises(ValueError) as cm:
            load_into_collections_from_zipfile(colls, strm)
        exc_msg = 'Could not infer MIME type'
        self.assert_true(str(cm.exception).startswith(exc_msg))

    def test_load_from_zipfile_filename_not_found(self):
        strm = BytesIO()
        zipf = zipfile.ZipFile(strm, 'w')
        zipf.writestr('foo.foo', '')
        zipf.close()
        colls = [get_root_collection(IMyEntity)]
        load_into_collections_from_zipfile(colls, strm)
        self.assert_equal(len(colls[0]), 0)


class ZipResourceIoTestCaseRdb(RdbTestCaseMixin, _ZipResourceIoTestCaseBase):
    config_file_name = 'configure.zcml'


class StreamResourceIoTestCase(_ResourceIoTestCaseBase):
    config_file_name = 'configure_no_rdb.zcml'
    def test_dump_no_content_type(self):
        member = _make_test_entity_member()
        strm = NativeIO()
        dump_resource(member, strm)
        self.assert_true(strm.getvalue().startswith('"id",'))


class FileResourceIoTestCase(_ResourceIoTestCaseBase):
    config_file_name = 'configure_no_rdb.zcml'
    def _test_load(self, load_func, fn_func, is_into):
        member = _make_test_entity_member()
        tmp_dir = tempfile.mkdtemp()
        try:
            dump_resource_to_files(member, directory=tmp_dir)
            file_names = glob.glob1(tmp_dir, "*.csv")
            self.assert_equal(len(file_names), 4)
            for ifc in [IMyEntityParent,
                        IMyEntity,
                        IMyEntityChild,
                        IMyEntityGrandchild]:
                coll_cls = get_collection_class(ifc)
                root_coll = get_root_collection(ifc)
                file_name = get_collection_filename(coll_cls)
                file_path = fn_func(os.path.join(tmp_dir, file_name))
                if not is_into:
                    coll = load_func(coll_cls, file_path)
                    self.assert_equal(len(coll), 1)
                    for mb in coll:
                        root_coll.add(mb)
                else:
                    load_func(root_coll, file_path)
                    self.assert_equal(len(root_coll), 1)
        finally:
            shutil.rmtree(tmp_dir)

    def test_load_from_invalid_file(self):
        coll_cls = get_collection_class(IMyEntity)
        with self.assert_raises(ValueError) as cm:
            dummy = \
              load_collection_from_file(coll_cls, 'my-entity-collection.foo')
        exc_msg = 'Could not infer MIME type'
        self.assert_true(str(cm.exception).startswith(exc_msg))

    def test_load_from_file(self):
        self._test_load(load_collection_from_file, lambda fn: fn, False)

    def test_load_from_stream(self):
        self._test_load(lambda rc, fn: load_collection_from_stream(rc, fn,
                                                                   CsvMime),
                        lambda fn: open(fn, 'rU'), False)

    def test_load_from_file_url(self):
        self._test_load(load_collection_from_url,
                        lambda fn: "file://%s" % fn, False)

    def test_load_from_file_url_into(self):
        self._test_load(load_into_collection_from_url,
                        lambda fn: "file://%s" % fn, True)

    def test_load_from_invalid_file_url(self):
        self.assert_raises(ValueError,
                           self._test_load,
                           load_collection_from_url,
                           lambda fn: "http://%s" % fn, False)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 18, 2011.
"""
from logging import StreamHandler
import logging
import random

from pyramid.compat import NativeIO

from everest.testing import Pep8CompliantTestCase
from everest.utils import BidirectionalLookup
from everest.utils import TruncatingFormatter
from everest.utils import WeakList
from everest.utils import WeakOrderedSet
from everest.utils import classproperty
from everest.utils import get_repository
from everest.utils import get_traceback
from everest.utils import id_generator
from everest.testing import EntityTestCase


__docformat__ = 'reStructuredText en'
__all__ = ['UtilsTestCase',
           ]


class UtilsTestCase(Pep8CompliantTestCase):

    def test_id_generator(self):
        # Initialize with start value.
        idgen = id_generator(3)
        self.assert_equal(next(idgen), 3)
        self.assert_equal(next(idgen), 4)
        # Push ID to higher value.
        idgen.send(5)
        self.assert_equal(next(idgen), 6)
        self.assert_equal(next(idgen), 7)
        # ID values must increase monotonically.
        self.assert_raises(ValueError, idgen.send, 6)

    def test_classproperty(self):
        class X(object):
            attr = 'myattr'
            @classproperty
            def clsprop(cls): # no self as first arg pylint: disable=E0213
                return cls.attr
        self.assert_equal(X.clsprop, X.attr)

    def test_get_traceback(self):
        try:
            raise RuntimeError('Something went wrong.')
        except RuntimeError:
            tb = get_traceback()
        self.assert_true(tb.startswith('Traceback (most recent call last)'))
        self.assert_true(tb.rstrip().endswith(
                                    'RuntimeError: Something went wrong.'))

    def test_bidirectional_lookup(self):
        bl = BidirectionalLookup(dict(a=1, b=2))
        self.assert_true(bl.has_left('a'))
        self.assert_true(bl.has_right(1))
        self.assert_false(bl.has_right('a'))
        self.assert_false(bl.has_left(1))
        self.assert_equal(bl.get_left('a'), 1)
        self.assert_equal(bl.get_right(1), 'a')
        self.assert_equal(bl.pop_left('a'), 1)
        self.assert_false(bl.has_left('a'))
        self.assert_false(bl.has_right(1))
        self.assert_equal(list(bl.left_keys()), ['b'])
        self.assert_equal(list(bl.right_keys()), [2])
        self.assert_equal(list(bl.left_values()), [2])
        self.assert_equal(list(bl.right_values()), ['b'])
        self.assert_equal(list(bl.left_items()), [('b', 2)])
        self.assert_equal(list(bl.right_items()), [(2, 'b')])
        self.assert_true('b' in bl)
        self.assert_true(2 in bl)
        self.assert_equal(bl['b'], 2)
        self.assert_equal(bl[2], 'b')
        self.assert_equal(bl.get('b'), 2)
        self.assert_equal(bl.get(2), 'b')
        self.assert_is_none(bl.get('not there'))
        self.assert_raises(ValueError, bl.__setitem__, 2, 'c')
        self.assert_raises(ValueError, bl.__setitem__, 0, 'b')
        self.assert_equal(bl.pop_left('not there', 5), 5)
        self.assert_equal(bl.pop_right('not there', 5), 5)
        bl['c'] = 2
        self.assert_true(2 in bl)
        self.assert_false('b' in bl)
        bl['c'] = 1
        self.assert_true(1 in bl)
        self.assert_false(2 in bl)
        bl['x'] = 3
        bl['y'] = 4
        self.assert_true('x' in bl)
        self.assert_true(4 in bl)
        del bl['x']
        del bl[4]
        self.assert_raises(KeyError, bl.__delitem__, 'x')
        self.assert_false('x' in bl)
        self.assert_false(4 in bl)
        bl.clear()
        self.assert_equal(len(bl.left_keys()), 0)
        self.assert_equal(len(bl.right_keys()), 0)

    def test_weaklist(self):
        class _X(object):
            def __init__(self):
                self.rank = random.randint(0, 20)
            def __lt__(self, other):
                return self.rank < other.rank
            def __le__(self, other):
                return self.rank <= other.rank
            def __eq__(self, other):
                return self.rank == other.rank
            def __ne__(self, other):
                return self.rank != other.rank
            def __gt__(self, other):
                return self.rank > other.rank
            def __ge__(self, other):
                return self.rank >= other.rank
            def __hash__(self):
                return id(self)
        class A(_X):
            pass
        class B(_X):
            pass
        class C(_X):
            pass
        obj_a = A()
        obj_b = B()
        weak_list = WeakList([obj_a, obj_b, obj_a, obj_b, obj_a, obj_b])
        # __len__() and deleting items through deleting the held items.
        self.assertEqual(len(weak_list), 6)
        del obj_b
        self.assertEqual(len(weak_list), 3)
        # "in" operation.
        self.assertEqual(obj_a in weak_list, True)
        # iteration.
        item = None
        count = 0
        for item in weak_list:
            count += 1
        self.assertEqual(count, 3)
        del item
        # count().
        self.assertEqual(weak_list.count(obj_a), 3)
        # pop().
        self.assertEqual(weak_list.pop() is obj_a, True)
        # insert() and __getitem__().
        obj_c = C()
        weak_list.insert(2, obj_c)
        self.assertEqual(weak_list[2] is obj_c, True)
        # __setitem__().
        weak_list[0] = obj_c
        self.assertEqual(isinstance(weak_list[0], C), True)
        # slicing.
        weak_list[1:2] = [obj_c]
        self.assertEqual(weak_list[1] is obj_c, True)
        self.assertEqual(weak_list[1:2], [obj_c])
        weak_list[slice(0, 1)] = [obj_c]
        self.assertEqual(weak_list[0] is obj_c, True)
        self.assertEqual(weak_list[slice(0, 1)], [obj_c])
        # index().
        self.assertEqual(weak_list.index(obj_c), 0)
        # .remove().
        weak_list.remove(obj_c)
        self.assertEqual(len(weak_list), 2)
        # extend().
        obj_c1 = C()
        obj_c2 = C()
        obj_c3 = C()
        weak_list.extend([obj_c1, obj_c2, obj_c3])
        self.assertEqual(len(weak_list), 5)
        # sort().
        item = None
        weak_list.sort()
        last_rank = -1
        for item in weak_list:
            self.assertEqual(last_rank <= item.rank, True)
            last_rank = item.rank
        del item
        # append.
        obj_c4 = C()
        weak_list.append(obj_c4)
        self.assert_true(weak_list[-1] is obj_c4)
        # add.
        self.assert_equal(len(weak_list + weak_list), 2 * len(weak_list))
        # clean up.
        del obj_c1, obj_c2, obj_c3, obj_c4
        del obj_a, obj_c
        self.assertEqual(len(weak_list), 0)
        with self.assert_raises(StopIteration):
            next(iter(weak_list))

    def test_weak_ordered_set(self):
        class MyObj(object):
            def __init__(self, value):
                self.value = value
            def __eq__(self, other):
                return self.value == other.value
            def __hash__(self):
                return hash(self.value)
        values = [MyObj(val) for val in range(5)]
        wos = WeakOrderedSet()
        self.assert_raises(KeyError, wos.pop)
        for value in values:
            wos.add(value)
        self.assert_equal(list(iter(wos)), values)
        self.assert_equal(list(reversed(wos)), values[::-1])
        self.assert_equal(len(wos), 5)
        self.assert_equal(wos.pop(), values.pop())
        self.assert_equal(len(wos), 4)
        other_wos = WeakOrderedSet(values)
        self.assert_equal(wos, other_wos)
        self.assert_equal(wos, values)

    def test_truncating_formatter(self):
        buf = NativeIO()
        logger = logging.Logger('test', logging.DEBUG)
        hdlr = StreamHandler(buf)
        hdlr.setFormatter(TruncatingFormatter())
        logger.addHandler(hdlr)
        logger.debug('%s', 'X' * 99, extra=dict(output_limit=100))
        self.assert_equal(len(buf.getvalue().strip()), 99)
        buf.seek(0)
        buf.truncate()
        logger.debug('%s', 'X' * 101, extra=dict(output_limit=100))
        self.assert_equal(len(buf.getvalue().strip()), 100)


class RepoManagerTestCase(EntityTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def test_get_repository(self):
        repo = get_repository('MEMORY')
        self.assert_equal(repo.name, 'MEMORY')
        # Default repo.
        repo = get_repository()
        self.assert_equal(repo.name, 'MEMORY')



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Mar 10, 2013.
"""
from everest.resources.staging import create_staging_collection
from everest.testing import ResourceTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.resources import MyEntityMember


__docformat__ = 'reStructuredText en'
__all__ = ['StagingCollectionTestCase',
           ]


class StagingCollectionTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def set_up(self):
        ResourceTestCase.set_up(self)
        self.coll = create_staging_collection(IMyEntity)

    def test_basics(self):
        foo = MyEntity(id=0)
        foo_mb = MyEntityMember.create_from_entity(foo)
        self.coll.add(foo_mb)
        agg = self.coll.get_aggregate()
        self.assert_true(agg.get_by_id(foo.id) is foo)
        self.assert_true(agg.get_by_slug(foo.slug) is foo)
        foo1 = MyEntity(id=0)
        txt = 'FROBNIC'
        foo1.text = txt
        agg.update(foo1)
        self.assert_equal(agg.get_by_id(foo.id).text, txt)
        self.assert_equal(len(list(agg.iterator())), 1)
        agg.remove(foo)
        self.assert_equal(len(list(agg.iterator())), 0)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 17, 2011.
"""
from pkg_resources import resource_filename # pylint: disable=E0611
from pyramid.compat import bytes_
from pyramid.compat import native_
from pyramid.testing import DummyRequest
import transaction

from everest.constants import RequestMethods
from everest.mime import CSV_MIME
from everest.mime import CsvMime
from everest.mime import XmlMime
from everest.renderers import RendererFactory
from everest.repositories.rdb.testing import RdbTestCaseMixin
from everest.resources.interfaces import IService
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_root_collection
from everest.resources.utils import get_service
from everest.resources.utils import resource_to_url
from everest.testing import FunctionalTestCase
from everest.testing import ResourceTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.testing import create_collection
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.interfaces import IFoo
from everest.tests.simple_app.resources import FooCollection
from everest.tests.simple_app.resources import FooMember
from everest.tests.simple_app.views import ExceptionPostCollectionView
from everest.tests.simple_app.views import ExceptionPutMemberView
from everest.tests.simple_app.views import UserMessagePostCollectionView
from everest.tests.simple_app.views import UserMessagePutMemberView
from everest.traversal import SuffixResourceTraverser
from everest.utils import get_repository_manager
from everest.views.getcollection import GetCollectionView
from everest.views.static import public_view
from everest.views.utils import accept_csv_only
import os


__docformat__ = 'reStructuredText en'
__all__ = ['BasicViewTestCase',
           'ClassicStyleConfiguredViewsTestCase',
           'ExceptionViewTestCase',
           'NewStyleConfiguredViewsTestCase',
           'PredicatedViewTestCase',
           'StaticViewTestCase',
           'WarningViewMemoryTestCase',
           'WarningViewRdbTestCase',
           'WarningWithExceptionViewTestCase',
           ]


class BasicViewTestCase(FunctionalTestCase):
    package_name = 'everest.tests.complete_app'
    ini_file_path = resource_filename('everest.tests.complete_app',
                                      'complete_app.ini')
    app_name = 'complete_app'
    path = '/my-entities/'

    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml('everest.tests.complete_app:configure_rpr.zcml')
        self.config.add_resource_view(IMyEntity,
                                      renderer='csv',
                                      request_method=RequestMethods.GET)
        self.config.add_member_view(IMyEntity,
                                    renderer='csv',
                                    request_method=RequestMethods.PUT)
        self.config.add_member_view(IMyEntity,
                                    renderer='csv',
                                    request_method=RequestMethods.PATCH)
        self.config.add_collection_view(IMyEntity,
                                        renderer='csv',
                                        request_method=RequestMethods.POST)
        self.config.add_collection_view(IMyEntityChild,
                                        renderer='csv',
                                        request_method=RequestMethods.POST)
        self.config.add_member_view(IMyEntity,
                                    renderer='csv',
                                    request_method=RequestMethods.DELETE)

    def test_get_collection_defaults(self):
        res = self.app.get(self.path, status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_slice_larger_max_size(self):
        create_collection()
        res = self.app.get(self.path, params=dict(size=10000), status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_invalid_slice_raises_error(self):
        create_collection()
        res = self.app.get(self.path, params=dict(size='foo'), status=500)
        self.assert_is_not_none(res)

    def test_get_collection_with_slice_size(self):
        create_collection()
        res = self.app.get(self.path, params=dict(size=1),
                           status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_slice_start(self):
        create_collection()
        res = self.app.get(self.path, params=dict(start=1, size=1),
                           status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_filter(self):
        create_collection()
        res = self.app.get(self.path, params=dict(q='id:equal-to:0'),
                           status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_order(self):
        create_collection()
        res = self.app.get(self.path, params=dict(sort='id:asc'),
                           status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_order_and_size(self):
        create_collection()
        res = self.app.get(self.path, params=dict(sort='id:asc', size=1),
                           status=200)
        self.assert_is_not_none(res)

    def test_get_collection_with_refs_options(self):
        # The links options are not processed by the renderers, so we need
        # a native everest view with a defined response MIME type.
        self.config.add_resource_view(IMyEntity,
                                      default_response_content_type=CsvMime,
                                      request_method=RequestMethods.GET)
        create_collection()
        res1 = self.app.get(self.path, params=dict(refs='parent:OFF'),
                           status=200)
        self.assert_is_not_none(res1)
        self.assert_equal(native_(res1.body).find(',"parent",'), -1)
        self.assert_equal(native_(res1.body).find(',"parent.id",'), -1)
        res2 = self.app.get(self.path, params=dict(refs='parent:INLINE'),
                           status=200)
        self.assert_is_not_none(res2)
        self.assert_equal(native_(res2.body).find(',"parent",'), -1)
        self.assert_not_equal(native_(res2.body).find(',"parent.id",'), -1)
        # Bogus refs parameters.
        self.app.get(self.path, params=dict(refs='parent:XXX'),
                     status=500)

    def test_get_member_default_content_type(self):
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        coll.create_member(ent)
        res = self.app.get("%s/0" % self.path, status=200)
        self.assert_is_not_none(res)

    def test_put_member(self):
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        mb = coll.create_member(ent)
        self.assert_equal(mb.__name__, '0')
        req_body = b'"id","text","number"\n0,"abc",2\n'
        res = self.app.put("%s/0" % self.path,
                           params=req_body,
                           content_type=CsvMime.mime_type_string,
                           status=200)
        self.assert_is_not_none(res)
        mb = next(iter(coll))
        self.assert_equal(mb.text, 'abc')
        self.assert_equal(mb.number, 2)
        req_body = b'"id","text","number"\n2,"abc",2\n'
        res = self.app.put("%s/0" % self.path,
                           params=req_body,
                           content_type=CsvMime.mime_type_string,
                           status=200)
        self.assert_equal(mb.id, 2)
        self.assert_true(res.headers['Location'].endswith('2/'))

    def test_patch_member(self):
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        mb = coll.create_member(ent)
        self.assert_equal(mb.__name__, '0')
        req_body = b'"number"\n2\n'
        res = self.app.patch("%s/0" % self.path,
                             params=req_body,
                             content_type=CsvMime.mime_type_string,
                             status=200)
        self.assert_is_not_none(res)
        mb = next(iter(coll))
        self.assert_equal(mb.number, 2)
        req_body = b'"id"\n2\n'
        res = self.app.patch("%s/0" % self.path,
                             params=req_body,
                             content_type=CsvMime.mime_type_string,
                             status=200)
        self.assert_equal(mb.id, 2)
        self.assert_true(res.headers['Location'].endswith('2/'))

    def test_patch_member_with_xml(self):
        self.config.add_member_view(IMyEntity,
                                    renderer='xml',
                                    request_method=RequestMethods.PATCH)
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        mb = coll.create_member(ent)
        req_body = \
            b'<tst:myentity xmlns:tst="http://xml.test.org/tests" id="0">' \
            b'    <tst:number>2</tst:number>' \
            b'</tst:myentity>'
        res = self.app.patch("%s/0" % self.path,
                             params=req_body,
                             content_type=XmlMime.mime_type_string,
                             status=200)
        self.assert_is_not_none(res)
        mb = next(iter(coll))
        self.assert_equal(mb.number, 2)

    def test_post_collection(self):
        new_id = 0
        req_body = b'"id","text","number"\n%d,"abc",2\n' % new_id
        res1 = self.app.post("%s" % self.path,
                             params=req_body,
                             content_type=CsvMime.mime_type_string,
                             status=201)
        self.assert_is_not_none(res1)
        coll = get_root_collection(IMyEntity)
        mb = coll[str(new_id)]
        self.assert_equal(mb.text, 'abc')

    def test_post_collection_no_id(self):
        req_body = b'"text","number"\n"abc",2\n'
        res = self.app.post("%s" % self.path,
                            params=req_body,
                            content_type=CsvMime.mime_type_string,
                            status=201)
        self.assert_is_not_none(res)
        self.assert_true(res.headers['Location'].endswith(self.path))
        self.assert_not_equal(native_(res.body).split(os.linesep)[1][:2],
                              '""')

    def test_post_nested_collection(self):
        mb, mb_url = self.__make_parent_and_link()
        child_coll = get_root_collection(IMyEntityChild)
        req_text = '"id","text","parent"\n0,"child","%s"\n' % mb_url
        res = self.app.post("%schildren" % mb_url,
                            params=bytes_(req_text, encoding='utf-8'),
                            content_type=CsvMime.mime_type_string,
                            status=201)
        self.assert_is_not_none(res)
        child_mb = child_coll['0']
        self.assert_equal(child_mb.text, 'child')
        self.assert_equal(child_mb.parent.id, mb.id)

    def test_post_nested_collection_no_parent(self):
        mb, mb_url = self.__make_parent_and_link()
        req_body = b'"id","text"\n0,"child"\n'
        res = self.app.post("%schildren" % mb_url,
                            params=req_body,
                            content_type=CsvMime.mime_type_string,
                            status=201)
        self.assert_is_not_none(res)
        child_coll = get_root_collection(IMyEntityChild)
        child_mb = child_coll['0']
        self.assert_equal(child_mb.text, 'child')
        self.assert_equal(child_mb.parent.id, mb.id)

    def test_delete_member(self):
        coll = create_collection()
        self.assert_equal(len(coll), 2)
        res = self.app.delete("%s/0" % self.path,
                              content_type=CsvMime.mime_type_string,
                              status=200)
        self.assert_is_not_none(res)
        self.assert_equal(len(coll), 1)
        # Second delete triggers 404.
        self.app.delete("%s/0" % self.path,
                        content_type=CsvMime.mime_type_string,
                        status=404)
        coll_cls = get_collection_class(IMyEntity)
        old_remove = coll_cls.__dict__.get('remove')
        def remove_with_exception(self): # pylint: disable=W0613
            raise RuntimeError()
        coll_cls.remove = remove_with_exception
        try:
            self.app.delete("%s/1" % self.path,
                            content_type=CsvMime.mime_type_string,
                            status=500)
        finally:
            if not old_remove is None:
                coll_cls.remove = old_remove

    def __make_parent_and_link(self):
        # FIXME: This is more elaborate than it should be - to make URL
        #        generation work, we have to manually set the parent of the
        #        root collection and create a dummy request.
        coll = get_root_collection(IMyEntity)
        svc = get_service()
        coll.__parent__ = svc
        ent = MyEntity(id=0)
        mb = coll.create_member(ent)
        # Make a dummy request.
        url = self.ini.get_app_url()
        req = DummyRequest(application_url=url, host_url=url,
                           path_url=url, url=url,
                           registry=self.config.registry)
        mb_url = resource_to_url(mb, request=req)
        return mb, mb_url


class MessagingViewTestCase(FunctionalTestCase):
    package_name = 'everest.tests.complete_app'
    ini_file_path = resource_filename('everest.tests.complete_app',
                                      'complete_app.ini')
    app_name = 'complete_app'
    path = '/my-entities/'

    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml('everest.tests.complete_app:configure_rpr.zcml')
        self.config.add_resource_view(IMyEntity,
                                      renderer='csv',
                                      request_method=RequestMethods.GET,
                                      enable_messaging=True)
        self.config.add_member_view(IMyEntity,
                                    renderer='csv',
                                    request_method=RequestMethods.PATCH,
                                    enable_messaging=False)

    def test_get_member_default_content_type(self):
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        coll.create_member(ent)
        res = self.app.get("%s/0" % self.path, status=200)
        self.assert_is_not_none(res)

    def test_patch_member(self):
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        mb = coll.create_member(ent)
        self.assert_equal(mb.__name__, '0')
        req_body = b'"number"\n2\n'
        res = self.app.patch("%s/0" % self.path,
                             params=req_body,
                             content_type=CsvMime.mime_type_string,
                             status=200)
        self.assert_is_not_none(res)


class PredicatedViewTestCase(FunctionalTestCase):
    package_name = 'everest.tests.complete_app'
    ini_file_path = resource_filename('everest.tests.complete_app',
                                      'complete_app.ini')
    app_name = 'complete_app'
    path = '/my-entities'
    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml('everest.tests.complete_app:configure_rpr.zcml')
        self.config.add_renderer('csv', RendererFactory)
        self.config.add_view(context=get_collection_class(IMyEntity),
                             view=GetCollectionView,
                             renderer='csv',
                             request_method=RequestMethods.GET,
                             custom_predicates=(accept_csv_only,))

    def test_csv_only(self):
        # Without accept header, we get a 404.
        self.app.get(self.path, status=404)
        self.app.get(self.path, headers=dict(accept=CSV_MIME), status=200)


class _ConfiguredViewsTestCase(FunctionalTestCase):
    views_config_file_name = None
    package_name = 'everest.tests.simple_app'
    ini_file_path = resource_filename('everest.tests.simple_app',
                                      'simple_app_views.ini')
    app_name = 'simple_app'
    path = '/foos'

    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml(self.views_config_file_name)
        coll = get_root_collection(IFoo)
        coll.create_member(FooEntity(id=0))
        transaction.commit()

    def test_with_suffix(self):
        self._test_with_view_name(
                            lambda path, suffix: "%s.%s" % (path, suffix))

    def test_with_at_at(self):
        self._test_with_view_name(
                            lambda path, suffix: "%s/@@%s" % (path, suffix))

    def test_custom_view_with_interface_raises_error(self):
        self.assert_raises(ValueError,
                           self.config.add_resource_view, IFoo,
                           view=lambda context, request: None)

    def _test_with_view_name(self, path_fn):
        # Use suffix traverser as default.
        self.config.add_traverser(SuffixResourceTraverser)
        for sfx, exp, end in (('csv', b'"id"', False),
                              ('json', b'[{"id": 0', False),
                              ('xml', b'</foos>', True)
                              ):
            res = self.app.get(path_fn(self.path, sfx), status=200)
            if not end:
                self.assert_equal(res.body[:len(exp)], exp)
            else:
                self.assert_equal(res.body.strip()[-len(exp):], exp)
        # Fail for non-existing collection.
        self.app.get('/bars.csv', status=404)


class ClassicStyleConfiguredViewsTestCase(_ConfiguredViewsTestCase):
    views_config_file_name = \
                    'everest.tests.simple_app:configure_views_classic.zcml'

    def test_default(self):
        # No default - triggers a 404.
        self.app.get(self.path, status=404)


class NewStyleConfiguredViewsTestCase(_ConfiguredViewsTestCase):
    views_config_file_name = \
                'everest.tests.simple_app:configure_views.zcml'

    def test_default(self):
        # New style views return the default_content_type.
        res = self.app.get(self.path, status=200)
        self.assert_true(res.body.startswith(b'<?xml'))

    def test_custom_view(self):
        TXT = b'my custom response body'
        def custom_view(context, request): # context unused pylint: disable=W0613
            request.response.body = TXT
            return request.response
        self.config.add_collection_view(IFoo, view=custom_view, name='custom')
        res = self.app.get('/foos/@@custom')
        self.assert_equal(res.body, TXT)

    def test_invalid_accept_header(self):
        self.app.get(self.path,
                     headers=dict(accept='application/foobar'),
                     status=406)

    def test_star_star_accept_header(self):
        self.app.get(self.path,
                     headers=dict(accept='*/*'),
                     status=200)

    def test_invalid_request_content_type(self):
        self.config.add_collection_view(IFoo,
                                        request_method=RequestMethods.POST)
        self.app.post(self.path,
                      params='foobar',
                      content_type='application/foobar',
                      status=415)

    def test_fake_put_view(self):
        self.config.add_member_view(IFoo,
                                    request_method=RequestMethods.FAKE_PUT)
        req_body = '"id"\n0'
        self.app.post("%s/0" % self.path,
                      params=req_body,
                      content_type=CsvMime.mime_type_string,
                      headers={'X-HTTP-Method-Override' : RequestMethods.PUT},
                      status=200)

    def test_fake_patch_view(self):
        self.config.add_member_view(IFoo,
                                    request_method=RequestMethods.FAKE_PATCH)
        req_body = '"id"\n0'
        self.app.post("%s/0" % self.path,
                      params=req_body,
                      content_type=CsvMime.mime_type_string,
                      headers={'X-HTTP-Method-Override' :
                                                    RequestMethods.PATCH},
                      status=200)

    def test_fake_delete_view(self):
        self.config.add_member_view(IFoo,
                                    request_method=RequestMethods.FAKE_DELETE)
        self.app.post("%s/0" % self.path,
                      headers=
                        {'X-HTTP-Method-Override' : RequestMethods.DELETE},
                      status=200)

    def test_add_collection_view_with_put_fails(self):
        with self.assert_raises(ValueError) as cm:
            self.config.add_collection_view(IFoo,
                                            request_method=RequestMethods.PUT)
        self.assert_true(str(cm.exception).startswith('Autodetection'))

    def test_add_member_view_with_post_fails(self):
        with self.assert_raises(ValueError) as cm:
            self.config.add_member_view(IFoo,
                                        request_method=RequestMethods.POST)
        self.assert_true(str(cm.exception).startswith('Autodetection'))


class GetCollectionViewTestCase(ResourceTestCase):
    package_name = 'everest.tests.simple_app'
    config_file_name = 'configure.zcml'

    def test_get_collection_view_with_size(self):
        coll = get_root_collection(IFoo)
        app_url = self.ini.get_app_url()
        path_url = 'http://0.0.0.0:6543/foos/'
        req = DummyRequest(application_url=app_url, host_url=app_url,
                           path_url=path_url,
                           url=path_url + '?size=10',
                           params=dict(size=10),
                           registry=self.config.registry,
                           accept=['*/*'])
        req.get_response = lambda exc: None
        view = GetCollectionView(coll, req)
        res = view()
        self.assert_is_not_none(res)
        self.assert_equal(view.context.slice.start, 0)
        self.assert_equal(view.context.slice.stop, 10)
        # Try again with size exceeding the allowed maximum limit (page size).
        req.params = dict(size=10000)
        req.url = path_url + '?size=10000'
        res = view()
        self.assert_is_not_none(res)
        self.assert_equal(view.context.slice.start, 0)
        self.assert_equal(view.context.slice.stop, FooCollection.max_limit)


class StaticViewTestCase(FunctionalTestCase):
    package_name = 'everest.tests.complete_app'
    ini_file_path = resource_filename('everest.tests.complete_app',
                                      'complete_app.ini')
    app_name = 'complete_app'
    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml(
                        'everest.tests.complete_app:configure_no_rdb.zcml')
        self.config.add_view(context=IService,
                             view=public_view,
                             name='public',
                             request_method=RequestMethods.GET)
        fn = resource_filename('everest.tests.complete_app', 'data/original')
        self.config.registry.settings['public_dir'] = fn

    def test_access_public_dir(self):
        self.app.get('/public/myentity-collection.csv', status=200)


class ExceptionViewTestCase(FunctionalTestCase):
    package_name = 'everest.tests.complete_app'
    ini_file_path = resource_filename('everest.tests.complete_app',
                                      'complete_app.ini')
    app_name = 'complete_app'
    path = '/my-entities'

    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml(
                        'everest.tests.complete_app:configure_no_rdb.zcml')
        self.config.add_member_view(IMyEntity,
                                    view=ExceptionPutMemberView,
                                    request_method=RequestMethods.PUT)
        self.config.add_collection_view(IMyEntity,
                                        view=ExceptionPostCollectionView,
                                        request_method=RequestMethods.POST)

    def test_put_member_raises_error(self):
        coll = get_root_collection(IMyEntity)
        ent = MyEntity(id=0)
        coll.create_member(ent)
        self.app.put("%s/0" % self.path,
                     params='dummy body',
                     status=500)

    def test_post_collection_raises_error(self):
        req_body = '"id","text","number"\n0,"abc",2\n'
        self.app.post("%s" % self.path,
                     params=req_body,
                     content_type=CsvMime.mime_type_string,
                     status=500)


class _WarningViewBaseTestCase(FunctionalTestCase):
    package_name = 'everest.tests.simple_app'
    ini_file_path = resource_filename('everest.tests.simple_app',
                                      'simple_app_views.ini')
    app_name = 'simple_app'
    path = '/foos'
    config_file_name = None

    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml(self.config_file_name)
        # We have to call this again to initialize the newly created SYSTEM
        # repo.
        repo_mgr = get_repository_manager()
        repo_mgr.initialize_all()
        self.config.add_collection_view(FooCollection,
                                        view=UserMessagePostCollectionView,
                                        request_method=RequestMethods.POST)
        self.config.add_member_view(FooMember,
                                    view=UserMessagePutMemberView,
                                    request_method=RequestMethods.PUT)

    def test_post_collection_empty_body(self):
        res = self.app.post(self.path, params='',
                            status=400)
        self.assert_false(res is None)

    def test_post_collection_warning_exception(self):
        # First POST - get back a 307.
        res1 = self.app.post(self.path, params='foo name',
                             status=307)
        body_text = native_(res1.body.rstrip(), encoding='utf-8')
        self.assert_true(body_text.endswith(
                                    UserMessagePostCollectionView.message))
        self.assert_true(res1.body.startswith(b'307 Temporary Redirect'))
        # Second POST to redirection location - get back a 201.
        resubmit_location1 = res1.headers['Location']
        res2 = self.app.post(resubmit_location1,
                             params='foo name',
                             status=201)
        self.assert_true(not res2 is None)
        # Third POST to same redirection location with different warning
        # message triggers a 307 again.
        old_msg = UserMessagePostCollectionView.message
        UserMessagePostCollectionView.message = old_msg[::-1]
        try:
            res3 = self.app.post(resubmit_location1,
                                 params='foo name',
                                 status=307)
            self.assert_true(res3.body.startswith(b'307 Temporary Redirect'))
            # Fourth POST to new redirection location - get back a 409 (since
            # the second POST from above went through).
            resubmit_location2 = res3.headers['Location']
            res4 = self.app.post(resubmit_location2,
                                 params='foo name',
                                 status=409)
            self.assert_true(not res4 is None)
        finally:
            UserMessagePostCollectionView.message = old_msg

    def test_post_collection_warning_exception_with_query_string(self):
        old_path = self.path
        self.path = '/foos?q=id=0'
        try:
            self.test_post_collection_warning_exception()
        finally:
            self.path = old_path

    def test_put_member_warning_exception(self):
        root = get_service()
        # Need to start the service manually - no request root has been set
        # yet.
        root.start()
        coll = root['foos']
        mb = FooMember(FooEntity(id=0))
        coll.add(mb)
        transaction.commit()
        path = '/'.join((self.path, '0'))
        # First PUT - get back a 307.
        res1 = self.app.put(path,
                            params='foo name',
                            status=307)
        self.assert_true(res1.body.startswith(b'307 Temporary Redirect'))
        # Second PUT to redirection location - get back a 200.
        resubmit_location1 = res1.headers['Location']
        res2 = self.app.put(resubmit_location1, params='foo name',
                            status=200)
        self.assert_true(not res2 is None)


class WarningViewMemoryTestCase(_WarningViewBaseTestCase):
    config_file_name = \
            'everest.tests.simple_app:configure_messaging_memory.zcml'


class WarningViewRdbTestCase(RdbTestCaseMixin, _WarningViewBaseTestCase):
    config_file_name = 'everest.tests.simple_app:configure_messaging_rdb.zcml'


class WarningWithExceptionViewTestCase(FunctionalTestCase):
    package_name = 'everest.tests.complete_app'
    ini_file_path = resource_filename('everest.tests.complete_app',
                                      'complete_app.ini')
    app_name = 'complete_app'
    path = '/my-entities'

    def set_up(self):
        FunctionalTestCase.set_up(self)
        self.config.load_zcml(
                        'everest.tests.complete_app:configure_no_rdb.zcml')
        self.config.add_view(context=get_collection_class(IMyEntity),
                             view=ExceptionPostCollectionView,
                             request_method=RequestMethods.POST)

    def test_post_collection_raises_error(self):
        req_body = '"id","text","number"\n0,"abc",2\n'
        self.app.post("%s" % self.path,
                     params=req_body,
                     content_type=CsvMime.mime_type_string,
                     status=500)



"""
Unit of work.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 16, 2013.
"""
from collections import defaultdict
from everest.repositories.state import ENTITY_STATUS
from everest.repositories.state import EntityState
from everest.utils import WeakOrderedSet

__docformat__ = 'reStructuredText en'
__all__ = ['UnitOfWork',
           ]


class UnitOfWork(object):
    """
    The Unit Of Work records object state changes for subsequent commit
    to a repository.

    Responsibilities:
     * Clone CLEAN entities upon registration;
     * Record entity state changes.
    """
    def __init__(self):
        self.__entity_set_map = defaultdict(WeakOrderedSet)

    def register_new(self, entity_class, entity):
        """
        Registers the given entity for the given class as NEW.

        :raises ValueError: If the given entity already holds state that was
          created by another Unit Of Work.
        """
        EntityState.manage(entity, self)
        EntityState.get_state(entity).status = ENTITY_STATUS.NEW
        self.__entity_set_map[entity_class].add(entity)

    def register_clean(self, entity_class, entity):
        """
        Registers the given entity for the given class as CLEAN.

        :returns: Cloned entity.
        """
        EntityState.manage(entity, self)
        EntityState.get_state(entity).status = ENTITY_STATUS.CLEAN
        self.__entity_set_map[entity_class].add(entity)

    def register_deleted(self, entity_class, entity):
        """
        Registers the given entity for the given class as DELETED.

        :raises ValueError: If the given entity already holds state that was
          created by another Unit Of Work.
        """
        EntityState.manage(entity, self)
        EntityState.get_state(entity).status = ENTITY_STATUS.DELETED
        self.__entity_set_map[entity_class].add(entity)

    def unregister(self, entity_class, entity):
        """
        Unregisters the given entity for the given class and discards its
        state information.
        """
        EntityState.release(entity, self)
        self.__entity_set_map[entity_class].remove(entity)

    def is_registered(self, entity):
        """
        Checks if the given entity is registered with this Unit Of Work.
        """
        return hasattr(entity, '__everest__') \
               and entity.__everest__.unit_of_work is self

    def is_marked_new(self, entity):
        """
        Checks if the given entity is marked with status NEW. Returns `False`
        if the entity has no state information.
        """
        try:
            result = EntityState.get_state(entity).status == ENTITY_STATUS.NEW
        except ValueError:
            result = False
        return result

    def is_marked_deleted(self, entity):
        """
        Checks if the given entity is marked with status DELETED. Returns
        `False` if the entity has no state information.
        """
        try:
            result = EntityState.get_state(entity).status \
                                                    == ENTITY_STATUS.DELETED
        except ValueError:
            result = False
        return result

    def is_marked_persisted(self, entity):
        """
        Checks if the flag indicating that the state for the given entity
        has been persisted is `True`. Returns `False` if the entity has no
        state information.
        """
        try:
            result = EntityState.get_state(entity).is_persisted
        except ValueError:
            result = False
        return result

    def is_marked_pending(self, entity):
        """
        Checks if the flag indicating that the state for the given entity
        has been persisted is `False`. Returns `False` if the entity has no
        state information.
        """
        try:
            result = not EntityState.get_state(entity).is_persisted
        except ValueError:
            result = False
        return result

    def mark_new(self, entity):
        """
        Marks the given entity as NEW.

        This is done when an entity is re-associated with a session after
        having been removed before.
        """
        EntityState.get_state(entity).status = ENTITY_STATUS.NEW

    def mark_clean(self, entity):
        """
        Marks the given entity as CLEAN.

        This is done when an entity is loaded fresh from the repository or
        after a commit.
        """
        state = EntityState.get_state(entity)
        state.status = ENTITY_STATUS.CLEAN
        state.is_persisted = True

    def mark_deleted(self, entity):
        """
        Marks the given entity as DELETED.

        :raises ValueError: If the given entity does not hold state.
        """
        EntityState.get_state(entity).status = ENTITY_STATUS.DELETED

    def mark_dirty(self, entity):
        """
        Marks the given entity for the given class as DIRTY.

        :raises ValueError: If the given entity does not hold state.
        """
        EntityState.get_state(entity).status = ENTITY_STATUS.DIRTY

    def mark_persisted(self, entity):
        """
        Sets the flag indicating if the state of the given entity has been
        persisted to `True`.

        :note: The persistency flag is orthogonal to the status flag.
        """
        EntityState.get_state(entity).is_persisted = True

    def mark_pending(self, entity):
        """
        Sets the flag indicating if the state of the given entity has been
        persisted to `False`.

        :note: The persistency flag is orthogonal to the status flag.
        """
        EntityState.get_state(entity).is_persisted = False

    def get_clean(self, entity_class=None):
        """
        Returns an iterator over all CLEAN entities in this Unit Of Work
        (optionally restricted to entities of the given class).
        """
        return self.__object_iterator(ENTITY_STATUS.CLEAN, entity_class)

    def get_new(self, entity_class=None):
        """
        Returns an iterator over all NEW entities in this Unit Of Work
        (optionally restricted to entities of the given class).
        """
        return self.__object_iterator(ENTITY_STATUS.NEW, entity_class)

    def get_deleted(self, entity_class=None):
        """
        Returns an iterator over all DELETED entities in this Unit Of Work
        (optionally restricted to entities of the given class).
        """
        return self.__object_iterator(ENTITY_STATUS.DELETED, entity_class)

    def get_dirty(self, entity_class=None):
        """
        Returns an iterator over all DIRTY entities in this Unit Of Work
        (optionally restricted to entities of the given class).
        """
        return self.__object_iterator(ENTITY_STATUS.DIRTY, entity_class)

    def iterator(self):
        """
        Returns an iterator over all entity states held by this Unit Of Work.
        """
        # FIXME: There is no dependency tracking; objects are iterated in
        #        random order.
        for ent_cls in list(self.__entity_set_map.keys()):
            for ent in self.__entity_set_map[ent_cls]:
                yield EntityState.get_state(ent)

    def reset(self):
        """
        Releases all entities held by this Unit Of Work (i.e., removes state
        information from all registered entities and clears the entity map).
        """
        for ents in self.__entity_set_map.values():
            for ent in ents:
                EntityState.release(ent, self)
        self.__entity_set_map.clear()

    def __object_iterator(self, status, ent_cls):
        if ent_cls is None:
            ent_clss = self.__entity_set_map.keys()
        else:
            ent_clss = [ent_cls]
        for ent_cls in ent_clss:
            for ent in self.__entity_set_map[ent_cls]:
                if EntityState.get_state(ent).status == status:
                    yield ent



"""
Resource attribute handling classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Dec 2, 2011.
"""
from collections import OrderedDict
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.resources.descriptors import attribute_base
from everest.resources.descriptors import collection_attribute \
    as collection_resource_attribute
from everest.resources.descriptors import member_attribute \
    as member_resource_attribute
from everest.resources.descriptors import terminal_attribute \
    as terminal_resource_attribute
from everest.resources.interfaces import IMemberResource
from everest.resources.utils import get_member_class
from functools import wraps
from pyramid.compat import iteritems_
from pyramid.compat import itervalues_
from zope.interface import implementedBy as implemented_by # pylint: disable=E0611,F0401
import functools

__docformat__ = 'reStructuredText en'
__all__ = ['MetaResourceAttributeCollector',
           'ResourceAttributeControllerMixin',
           'get_resource_class_attribute',
           'get_resource_class_attribute_iterator',
           'get_resource_class_attribute_names',
           'get_resource_class_attributes',
           'get_resource_class_collection_attribute_iterator',
           'get_resource_class_member_attribute_iterator',
           'get_resource_class_relationship_attribute_iterator',
           'get_resource_class_terminal_attribute_iterator',
           'is_resource_class_collection_attribute',
           'is_resource_class_member_attribute',
           'is_resource_class_resource_attribute',
           'is_resource_class_terminal_attribute',
           'resource_attributes_injector',
           ]


class MetaResourceAttributeCollector(type):
    """
    Meta class for member resource classes managing declared attributes.

    Extracts relevant information from the resource class descriptors for
    use e.g. in the representers.
    """

    def __init__(mcs, name, bases, class_dict):
        # Skip classes that are direct subclasses of the base mixin class.
        if name != 'ResourceAttributeControllerMixin' \
           and not ResourceAttributeControllerMixin in bases:
            dicts = []
            for cls in mcs.__mro__[::-1]:
                if cls in ResourceAttributeControllerMixin.__mro__:
                    continue
                dicts.append(cls.__dict__)
            attr_map = mcs.__collect_attributes(dicts)
            # Store in class namespace.
            mcs.__everest_attributes__ = attr_map
        type.__init__(mcs, name, bases, class_dict)

    def __collect_attributes(mcs, dicts):
        # Loop over the namespace of the given resource class and its base
        # classes looking for descriptors inheriting from
        # :class:`everest.resources.descriptors.attribute_base`.
        descr_map = {}
        for base_cls_namespace in dicts:
            for descr_name, descr in iteritems_(base_cls_namespace):
                if isinstance(descr, attribute_base):
                    descr_map[descr_name] = descr
        # Order by descriptor index (=sequence in which they were declared).
        ordered_descr_map = OrderedDict()
        cmp_fnc = lambda item1, item2: (item1[1].index > item2[1].index) - \
                                       (item1[1].index < item2[1].index)
        for item in sorted(descr_map.items(),
                           key=functools.cmp_to_key(cmp_fnc)):
            name, descr = item
            if not type(descr) in (terminal_resource_attribute,
                                   member_resource_attribute,
                                   collection_resource_attribute):
                raise TypeError('Unknown resource attribute type "%s".'
                                % type(descr))
            ordered_descr_map[name] = descr
            # It would be awkward to repeat the resource attribute name
            # in the parameters to the descriptor, so we set it manually
            # here.
            descr.resource_attr = name
        return ordered_descr_map


# PY3 compatible way of using the metaclass (__metaclass__ does not work).
ResourceAttributeControllerMixin = MetaResourceAttributeCollector(
                                        'ResourceAttributeControllerMixin',
                                        (object,), {})
# This is populated by the meta class.
ResourceAttributeControllerMixin.__everest_attributes__ = None


def arg_to_member_class(func):
    @wraps(func)
    def wrap(rc, *args):
        if isinstance(rc, type) and IMemberResource in implemented_by(rc):
            mb_cls = rc
        else:
            mb_cls = get_member_class(rc)
        return func(mb_cls, *args)
    return wrap


def is_resource_class_terminal_attribute(rc, attr_name):
    """
    Checks if the given attribute name is a terminal attribute of the given
    registered resource.
    """
    attr = get_resource_class_attribute(rc, attr_name)
    return attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL


def is_resource_class_member_attribute(rc, attr_name):
    """
    Checks if the given attribute name is a member attribute of the given
    registered resource.
    """
    attr = get_resource_class_attribute(rc, attr_name)
    return attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER


def is_resource_class_collection_attribute(rc, attr_name):
    """
    Checks if the given attribute name is a collection attribute of the given
    registered resource.
    """
    attr = get_resource_class_attribute(rc, attr_name)
    return attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION


def is_resource_class_resource_attribute(rc, attr_name):
    """
    Checks if the given attribute name is a resource attribute (i.e., either
    a member or a collection attribute) of the given registered resource.
    """
    attr = get_resource_class_attribute(rc, attr_name)
    return attr != RESOURCE_ATTRIBUTE_KINDS.TERMINAL


@arg_to_member_class
def get_resource_class_attribute(rc, name):
    """
    Returns the specified resource class attribute from the map of all
    collected attributes for the given registered resource or `None`, if the
    attribute could not be found.
    """
    return rc.__everest_attributes__.get(name)


@arg_to_member_class
def get_resource_class_attributes(rc):
    """
    Returns a dictionary mapping the attribute names of the given
    registered resource to :class:`ResourceAttribute` instances.
    """
    return rc.__everest_attributes__


@arg_to_member_class
def get_resource_class_attribute_names(rc):
    """
    Returns all attribute names of the given registered resource.
    """
    return rc.__everest_attributes__.keys()


@arg_to_member_class
def get_resource_class_attribute_iterator(rc):
    """
    Returns an iterator over all attributes in the given registered resource.
    """
    for attr in rc.__everest_attributes__.values():
        yield attr

@arg_to_member_class
def get_resource_class_terminal_attribute_iterator(rc):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(rc.__everest_attributes__):
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
            yield attr


@arg_to_member_class
def get_resource_class_relationship_attribute_iterator(rc):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(rc.__everest_attributes__):
        if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
            yield attr


@arg_to_member_class
def get_resource_class_member_attribute_iterator(rc):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(rc.__everest_attributes__):
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
            yield attr


@arg_to_member_class
def get_resource_class_collection_attribute_iterator(rc):
    """
    Returns an iterator over all terminal attributes in the given registered
    resource.
    """
    for attr in itervalues_(rc.__everest_attributes__):
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
            yield attr


class resource_attributes_injector(object):
    """
    Attribute injector.

    This is used to inject the resource attribute descriptor map (using the
    entity attribute name as key) as "__everest_atttributes__" attribute into
    the namespace of the corresponding entity class.
    """
    def __get__(self, dummy, entity_class):
        mb_cls = get_member_class(entity_class)
        attr_map = OrderedDict()
        for rc_attr in itervalues_(mb_cls.__everest_attributes__):
            attr_map[rc_attr.entity_attr] = rc_attr
        # This replaces the injector.
        entity_class.__everest_attributes__ = attr_map
        return attr_map



"""
Querying functionality for the rdb repository.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.base import Entity
from everest.exceptions import MultipleResultsException
from everest.exceptions import NoResultsException
from everest.querying.base import EXPRESSION_KINDS
from everest.querying.filtering import RepositoryFilterSpecificationVisitor
from everest.querying.interfaces import IFilterSpecificationVisitor
from everest.querying.interfaces import IOrderSpecificationVisitor
from everest.querying.ordering import OrderSpecificationVisitor
from everest.querying.ordering import RepositoryOrderSpecificationVisitor
from everest.querying.specifications import order
from everest.repositories.rdb.utils import OrmAttributeInspector
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IResource
from everest.utils import get_order_specification_visitor
from functools import reduce as func_reduce
from sqlalchemy import and_ as sqlalchemy_and
from sqlalchemy import not_ as sqlalchemy_not
from sqlalchemy import or_ as sqlalchemy_or
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.query import Query as SaQuery
from sqlalchemy.sql.expression import ClauseList
from sqlalchemy.sql.expression import func
from sqlalchemy.sql.expression import over
from zope.interface import implementer # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['OptimizedCountingQuery',
           'OrderClauseList',
           'SimpleCountingQuery',
           'SqlFilterSpecificationVisitor',
           'SqlOrderSpecificationVisitor',
           ]


@implementer(IFilterSpecificationVisitor)
class SqlFilterSpecificationVisitor(RepositoryFilterSpecificationVisitor):
    """
    Filter specification visitor implementation for the RDB repository
    (builds a SQL expression).
    """
    def __init__(self, entity_class, custom_clause_factories=None):
        """
        Constructs a SqlFilterSpecificationVisitor

        :param entity_class: an entity class that is mapped with SQLAlchemy
        :param custom_clause_factories: a map containing custom clause factory
          functions for selected (attribute name, operator) combinations.
        """
        RepositoryFilterSpecificationVisitor.__init__(self, entity_class)
        if custom_clause_factories is None:
            custom_clause_factories = {}
        self.__custom_clause_factories = custom_clause_factories

    def visit_nullary(self, spec):
        key = (spec.attr_name, spec.operator.name)
        if key in self.__custom_clause_factories:
            self._push(self.__custom_clause_factories[key](spec.attr_value))
        else:
            RepositoryFilterSpecificationVisitor.visit_nullary(self, spec)

    def _starts_with_op(self, spec):
        return self.__build(spec.attr_name, 'startswith', spec.attr_value)

    def _ends_with_op(self, spec):
        return self.__build(spec.attr_name, 'endswith', spec.attr_value)

    def _contains_op(self, spec):
        return self.__build(spec.attr_name, 'contains', spec.attr_value)

    def _contained_op(self, spec):
        if ICollectionResource.providedBy(spec.attr_value): # pylint:disable=E1101
            # FIXME: This is a hack that allows us to query for containment
            #        of a member in an arbitrary collection (not supported
            #        by SQLAlchemy yet).
            spec.attr_name = spec.attr_name + '.id'
            spec.attr_value = [rc.id for rc in spec.attr_value]
        return self.__build(spec.attr_name, 'in_', spec.attr_value)

    def _equal_to_op(self, spec):
        return self.__build(spec.attr_name, '__eq__', spec.attr_value)

    def _less_than_op(self, spec):
        return self.__build(spec.attr_name, '__lt__', spec.attr_value)

    def _less_than_or_equal_to_op(self, spec):
        return self.__build(spec.attr_name, '__le__', spec.attr_value)

    def _greater_than_op(self, spec):
        return self.__build(spec.attr_name, '__gt__', spec.attr_value)

    def _greater_than_or_equal_to_op(self, spec):
        return self.__build(spec.attr_name, '__ge__', spec.attr_value)

    def _in_range_op(self, spec):
        from_value, to_value = spec.attr_value
        return self.__build(spec.attr_name, 'between', from_value, to_value)

    def _conjunction_op(self, spec, *expressions):
        return sqlalchemy_and(*expressions)

    def _disjunction_op(self, spec, *expressions):
        return sqlalchemy_or(*expressions)

    def _negation_op(self, spec, expression):
        return sqlalchemy_not(expression)

    def __build(self, attribute_name, sql_op, *values):
        # Builds an SQL expression from the given (possibly dotted)
        # attribute name, SQL operation name, and values.
        exprs = []
        infos = OrmAttributeInspector.inspect(self._entity_class,
                                              attribute_name)
        count = len(infos)
        for idx, info in enumerate(infos):
            kind, entity_attr = info
            if idx == count - 1:
                #
                args = \
                    [val.get_entity() if IResource.providedBy(val) else val # pylint: disable=E1101
                     for val in values]
                expr = getattr(entity_attr, sql_op)(*args)
            elif kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
                expr = entity_attr.has
                exprs.insert(0, expr)
            elif kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION:
                expr = entity_attr.any
                exprs.insert(0, expr)
        return func_reduce(lambda g, h: h(g), exprs, expr)


class OrderClauseList(ClauseList):
    """
    Custom clause list for ORDER BY clauses.

    Suppresses the grouping parentheses which would trigger a syntax error.
    """
    def self_group(self, against=None):
        return self


@implementer(IOrderSpecificationVisitor)
class SqlOrderSpecificationVisitor(RepositoryOrderSpecificationVisitor):
    """
    Order specification visitor implementation for the rdb repository
    (builds a SQL expression).
    """
    def __init__(self, entity_class, custom_join_clauses=None):
        """
        Constructs a SqlOrderSpecificationVisitor

        :param klass: a class that is mapped to a selectable using SQLAlchemy
        """
        RepositoryOrderSpecificationVisitor.__init__(self, entity_class)
        if custom_join_clauses is None:
            custom_join_clauses = {}
        self.__custom_join_clauses = custom_join_clauses
        self.__joins = set()

    def visit_nullary(self, spec):
        OrderSpecificationVisitor.visit_nullary(self, spec)
        if spec.attr_name in self.__custom_join_clauses:
            self.__joins = set(self.__custom_join_clauses[spec.attr_name])

    def order_query(self, query):
        for join_expr in self.__joins:
            # FIXME: only join when needed here.
            query = query.outerjoin(join_expr)
        return query.order(self.expression)

    def _conjunction_op(self, spec, *expressions):
        clauses = []
        for expr in expressions:
            if isinstance(expr, ClauseList):
                clauses.extend(expr.clauses)
            else:
                clauses.append(expr)
        return OrderClauseList(*clauses)

    def _asc_op(self, spec):
        return self.__build(spec.attr_name, 'asc')

    def _desc_op(self, spec):
        return self.__build(spec.attr_name, 'desc')

    def __build(self, attribute_name, sql_op):
        expr = None
        infos = OrmAttributeInspector.inspect(self._entity_class,
                                              attribute_name)
        count = len(infos)
        for idx, info in enumerate(infos):
            kind, entity_attr = info
            if idx == count - 1:
                expr = getattr(entity_attr, sql_op)()
            elif kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                # FIXME: Avoid adding multiple attrs with the same target here.
                self.__joins.add(entity_attr)
        return expr


class Query(SaQuery):
    def __init__(self, entities, session, **kw):
        SaQuery.__init__(self, entities, session, **kw)
        ent_cls = entities[0]
        if isinstance(ent_cls, type) and issubclass(ent_cls, Entity):
            self._entity_class = ent_cls
        else: # just for compatibility pragma: no cover
            self._entity_class = None

    def order(self, order_expression):
        return SaQuery.order_by(self, order_expression)

    def order_by(self, *args):
        spec = order(*args)
        vst_cls = get_order_specification_visitor(EXPRESSION_KINDS.SQL)
        vst = vst_cls(self._entity_class)
        spec.accept(vst)
        return vst.order_query(self)


class CountingQuery(Query):
    def __init__(self, entities, session, **kw):
        Query.__init__(self, entities, session, **kw)
        self.__count = None
        self.__data = None

    def __iter__(self):
        if self.__data is None:
            self.__count, self.__data = self._load()
        return iter(self.__data)

    def count(self):
        if self.__count is None:
            self.__count, self.__data = self._load()
        return self.__count

    def one(self):
        # Overwritten so we can translate exceptions.
        try:
            return Query.one(self)
        except NoResultFound:
            raise NoResultsException('No results found when exactly one '
                                     'was expected.')
        except MultipleResultsFound:
            raise MultipleResultsException('More than one result found '
                                           'where exactly one was expected.')

    def _load(self):
        raise NotImplementedError('Abstract method.')

    def _clone(self):
        clone = Query._clone(self)
        # pylint: disable=W0212
        clone.__data = None
        clone.__count = None
        # pylint: enable=W0212
        return clone


class SimpleCountingQuery(CountingQuery):
    """
    Non-optimized counting query.
    """
    def _load(self):
        count_query = self.limit(None).offset(None)
        # Avoid circular calls to _load by "downcasting" the new query.
        count_query.__class__ = Query
        count = count_query.count()
        return count, list(Query.__iter__(self))


class OptimizedCountingQuery(CountingQuery): # pragma: no cover
    """
    Optimized counting query.

    This uses the OVER windowing SQL statement (on backends that have it) to
    reduce the loading time of everest collections by a factor of almost two
    (by virtue of avoiding an extra database roundtrip to obtain the count
    of the filtered collection required for paging).
    """
    def _load(self):
        query = self.add_columns(over(func.count(1)).label('_count'))
        res = [tup[0] for tup in Query.__iter__(query)]
        if len(res) > 0:
            count = tup._count # pylint:disable-msg=W0212,W0631
        else:
            count = 0
        return count, res



"""
CSV representers.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 19, 2011.
"""
from __future__ import absolute_import # Makes the import below absolute

from collections import OrderedDict
from csv import Dialect
from csv import QUOTE_NONNUMERIC
from csv import register_dialect
from csv import writer
import datetime
from itertools import product

from pyramid.compat import NativeIO
from pyramid.compat import PY3
from pyramid.compat import binary_type
from pyramid.compat import bytes_
from pyramid.compat import iteritems_
from pyramid.compat import string_types
from pyramid.compat import text_
from pyramid.compat import text_type

from everest.compat import CsvDictReader
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.constants import RESOURCE_KINDS
from everest.mime import CsvMime
from everest.representers.attributes import MappedAttributeKey
from everest.representers.base import MappingResourceRepresenter
from everest.representers.base import RepresentationGenerator
from everest.representers.base import RepresentationParser
from everest.representers.config import RepresenterConfiguration
from everest.representers.converters import BooleanConverter
from everest.representers.converters import ConverterRegistry
from everest.representers.converters import DateTimeConverter
from everest.representers.converters import NoOpConverter
from everest.representers.dataelements import SimpleCollectionDataElement
from everest.representers.dataelements import SimpleLinkedDataElement
from everest.representers.dataelements import SimpleMemberDataElement
from everest.representers.interfaces import IRepresentationConverter
from everest.representers.mapping import SimpleMappingRegistry
from everest.representers.traversal import DataElementTreeTraverser
from everest.representers.traversal import ResourceDataVisitor
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from everest.resources.utils import is_resource_url
from everest.resources.utils import provides_member_resource
from zope.interface import provider # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['CsvCollectionDataElement',
           'CsvConverterRegistry',
           'CsvData',
           'CsvDataElementTreeVisitor',
           'CsvIntConverter',
           'CsvLinkedDataElement',
           'CsvMappingRegistry',
           'CsvMemberDataElement',
           'CsvRepresentationGenerator',
           'CsvRepresentationParser',
           'CsvRepresenterConfiguration',
           'CsvResourceRepresenter',
           ]


class _DefaultCsvDialect(Dialect): # ignore no __init__ pylint: disable=W0232
    """
    Default dialect to use when exporting resources to CSV.
    """
    delimiter = ','
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\n'
    quoting = QUOTE_NONNUMERIC
register_dialect('export', _DefaultCsvDialect)
register_dialect('import', _DefaultCsvDialect)


class CsvConverterRegistry(ConverterRegistry):
    pass


@provider(IRepresentationConverter)
class CsvIntConverter(object):
    """
    Specialized converter coping with the CSV reader's unfortunate habit
    to convert integers to floats upon reading.
    """

    @classmethod
    def from_representation(cls, value):
        if isinstance(value, float):
            value = int(value)
        return value

    @classmethod
    def to_representation(cls, value):
        return value


if not PY3: # pragma: no cover
    @provider(IRepresentationConverter)
    class CsvTextConverter(object):

        @classmethod
        def from_representation(cls, value):
            return text_(value, encoding='utf-8')

        @classmethod
        def to_representation(cls, value):
            return bytes_(value, encoding='utf-8')

    CsvConverterRegistry.register(text_type, CsvTextConverter)
else: # pragma: no cover
    @provider(IRepresentationConverter)
    class CsvBytesConverter(object):

        @classmethod
        def from_representation(cls, value):
            return bytes_(value, encoding='utf-8')

        @classmethod
        def to_representation(cls, value):
            return text_(value, encoding='utf-8')

    CsvConverterRegistry.register(binary_type, CsvBytesConverter)


CsvConverterRegistry.register(datetime.datetime, DateTimeConverter)
CsvConverterRegistry.register(bool, BooleanConverter)
CsvConverterRegistry.register(int, CsvIntConverter)
CsvConverterRegistry.register(float, NoOpConverter)


class CsvRepresentationParser(RepresentationParser):
    """
    Parser for CSV representations.

    The tabular structure of the CSV format makes it difficult to transport
    nested (tree-like) data structures with it. The simplest way to solve
    this problem is to resort to links (URLs) for specifying nested resources
    which implies that the referenced resources have to be created in
    advance.

    Since it is quite a common use case to create a resource and its
    immediate children from one representation (i.e., in a single REST call),
    the CSV parser also supports explicit specification of nested member
    attributes through separate CSV fields and specification of nested
    collection member attributes through separate
    CSV fields and multiple rows (i.e., each row is specifying a member in
    the nested collection while all other field values remain unchanged).
    Nested collection members are allocated to their enclosing member either
    by the member ID (if specified in a column named "id") or by the
    combined values of all remaining fields (sorted by field name).

    :note: The CSV column (field) names have to be mapped uniquely to
      (nested) attribute representation names.
    :note: Polymorphic nested resources may not be mapped correctly.
    """
    class _CollectionData(object):
        def __init__(self, collection_class, attributes, attribute_key):
            self.collection_class = collection_class
            self.attribute_key = attribute_key
            self.__attribute_names = \
                        set([attr.repr_name for attr in attributes])
            self.__data = {}

        def has(self, key):
            return key in self.__data

        def get(self, key):
            return self.__data.get(key)

        def set(self, key, collection_data_element):
            self.__data[key] = collection_data_element

        def make_key(self, row_data):
            if "id" in row_data.keys():
                key = row_data['id']
            else:
                # FIXME: This needs testing.
                key = tuple([val for (key, val) in sorted(row_data.items()) # pragma: no cover
                             if not key in self.__attribute_names])
            return key

    def __init__(self, stream, resource_class, mapping):
        RepresentationParser.__init__(self, stream, resource_class, mapping)
        # Helper object for collecting member data for a nested collection.
        self.__coll_data = None
        # List of field names and copy of row data (first row only).
        self.__first_row_field_names = None
        self.__first_row_data = None
        # Flag indicating state "in first row".
        self.__is_first_row = True
        # Key used to detect repeating rows for nested collection members.
        self.__row_data_key = None

    def run(self):
        csv_rdr = CsvDictReader(self._stream,
                                dialect=self.get_option('dialect'))
        is_member_rpr = provides_member_resource(self._resource_class)
        if is_member_rpr:
            coll_data_el = None
        else:
            coll_data_el = self._mapping.create_data_element()
        for row_data in csv_rdr:
            if self.__is_first_row:
                self.__first_row_field_names = set(csv_rdr.fieldnames)
                self.__first_row_data = row_data.copy()
            if not self.__coll_data is None:
                # We need to generate the row data key now because we
                # get attribute values destructively from the row_data.
                self.__row_data_key = self.__coll_data.make_key(row_data)
            mb_data_el = self.__process_row(row_data, self._resource_class,
                                            MappedAttributeKey(()))
            if self.__is_first_row:
                self.__is_first_row = False
                if len(self.__first_row_field_names) > 0:
                    raise ValueError('Invalid field name(s): %s'
                                     % ','.join(self.__first_row_field_names))
            if None in row_data.keys():
                raise ValueError('Invalid row length.')
            if not coll_data_el is None:
                # The member data element will be None for all but the first
                # member of nested collection resources.
                if not mb_data_el is None:
                    coll_data_el.add_member(mb_data_el)
        if is_member_rpr:
            result_data_el = mb_data_el
        else:
            result_data_el = coll_data_el
        return result_data_el

    def __process_row(self, row_data, mapped_class, attribute_key):
        is_repeating_row = len(attribute_key) == 0 \
                           and not self.__coll_data is None \
                           and self.__coll_data.has(self.__row_data_key)
        if is_repeating_row:
            self.__process_row(row_data,
                               self.__coll_data.collection_class,
                               self.__coll_data.attribute_key)
            new_data_el = None
        else:
            mb_cls = get_member_class(mapped_class)
            new_data_el = \
                    self._mapping.create_data_element(mapped_class=mb_cls)
            for attr in self._mapping.terminal_attribute_iterator(
                                                mapped_class, attribute_key):
                self.__process_attr(attr, attribute_key, new_data_el,
                                    row_data)
            for attr in self._mapping.nonterminal_attribute_iterator(
                                                mapped_class, attribute_key):
                self.__process_attr(attr, attribute_key, new_data_el,
                                    row_data)
            if not self.__coll_data is None \
               and mapped_class == self.__coll_data.collection_class:
                if not self.__coll_data.has(self.__row_data_key):
                    coll_data_el = \
                        self._mapping.create_data_element(mapped_class=
                                                                mapped_class)
                    self.__coll_data.set(self.__row_data_key, coll_data_el)
                else:
                    coll_data_el = self.__coll_data.get(self.__row_data_key)
                coll_data_el.add_member(new_data_el)
                new_data_el = coll_data_el
        return new_data_el

    def __process_attr(self, attribute, attribute_key, data_el, row_data):
        try:
            attribute_value = row_data.pop(attribute.repr_name)
        except KeyError:
            attribute_value = None
            has_attribute = False
        else:
            has_attribute = True
        # In the first row, we check for extra field names by removing all
        # fields we found from the set of all field names.
        if self.__is_first_row:
            self.__first_row_field_names.discard(attribute.repr_name)
        if attribute.should_ignore(attribute_key):
            if not attribute_value in (None, ''):
                raise ValueError('Value for attribute "%s" found '
                                 'which is configured to be ignored.'
                                 % attribute.repr_name)
        else:
            if attribute.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                if not attribute_value is None:
                    data_el.set_terminal_converted(attribute, attribute_value)
            else:
                # FIXME: It is peculiar to treat the empty string as None
                #        here. However, this seems to be the way the csv
                #        module does it.
                if not attribute_value in (None, ''):
                    link_data_el = \
                            self.__process_link(attribute_value, attribute)
                    data_el.set_nested(attribute, link_data_el)
                elif not has_attribute and len(row_data) > 0:
                    nested_attr_key = attribute_key + (attribute,)
                    # We recursively look for nested resource attributes in
                    # other fields.
                    if attribute.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
                        # For polymorphic classes, this lookup will only work
                        # if a representer (and a mapping) was initialized
                        # for each derived class.
                        nested_rc_cls = get_member_class(attribute.value_type)
                    else: # collection attribute.
                        nested_rc_cls = \
                                    get_collection_class(attribute.value_type)
                        if self.__coll_data is None:
                            self.__coll_data = \
                                self.__make_collection_data(nested_rc_cls,
                                                            nested_attr_key)
                            if self.__is_first_row:
                                self.__row_data_key = \
                                    self.__coll_data.make_key(
                                                        self.__first_row_data)
                        elif self.__coll_data.collection_class \
                                                         != nested_rc_cls:
                            raise ValueError('All but one nested collection '
                                             'resource attributes have to '
                                             'be provided as links.')
                    nested_data_el = \
                        self.__process_row(row_data, nested_rc_cls,
                                           nested_attr_key)
                    if attribute.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
                        if len(nested_data_el.data) > 0:
                            data_el.set_nested(attribute, nested_data_el)
                    elif len(nested_data_el) > 0:
                        data_el.set_nested(attribute, nested_data_el)

    def __make_collection_data(self, collection_class, attribute_key):
        attrs = self._mapping.attribute_iterator(collection_class,
                                                 attribute_key)
        return self._CollectionData(collection_class, attrs, attribute_key)

    def __is_link(self, value):
        return isinstance(value, string_types) and is_resource_url(value)

    def __process_link(self, link, attr):
        if not self.__is_link(link):
            raise ValueError('Value for nested attribute "%s" '
                             'is not a link.' % attr.repr_name)
        if attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER:
            kind = RESOURCE_KINDS.MEMBER
            rc_cls = get_member_class(attr.value_type)
        else:
            kind = RESOURCE_KINDS.COLLECTION
            rc_cls = get_collection_class(attr.value_type)
        return self._mapping.create_linked_data_element(link,
                                                        kind,
                                                        relation=
                                                            rc_cls.relation,
                                                        title=rc_cls.title)


#class CsvRepresentationParser(RepresentationParser):
#    """
#    Parser converting CSV representations of resources into a data element.
#
#    :note: Nested resources have to be provided as links (i.e., there is no
#           support for recursive data element tree building).
#    """
#
#    def run(self):
#        mp_reg = get_mapping_registry(CsvMime)
#        is_member_rpr = provides_member_resource(self._resource_class)
#        if is_member_rpr:
#            member_cls = self._resource_class
#            result_data_el = None
#        else:
#            # Collection resource: Create a wrapping collection data element.
#            member_cls = get_member_class(self._resource_class)
#            coll_mp = mp_reg.find_or_create_mapping(self._resource_class)
#            coll_data_el = coll_mp.create_data_element()
#            result_data_el = coll_data_el
#        mb_mp = mp_reg.find_or_create_mapping(member_cls)
#        csv_rdr = reader(self._stream, self.get_option('dialect'))
#        attrs = mb_mp.get_attribute_map()
#        header = None
#        for row in csv_rdr:
#            mb_data_el = mb_mp.create_data_element()
#            if header is None:
#                # Check if the header is valid.
#                attr_names = attrs.keys()
#                header = row
#                for attr in header:
#                    if not attr in attr_names:
#                        raise ValueError('Invalid field "%s" in CSV input '
#                                         'detected.' % attr)
#                continue
#            if len(row) != len(header):
#                raise ValueError("Invalid row length (found: %s, expected: "
#                                 "%s)." % (len(row), len(header)))
#            for csv_attr, value in zip(header, row):
#                if value == '':
#                    value = None
#                attr = attrs[csv_attr]
#                if is_resource_url(value):
#                    link = CsvLinkedDataElement.create(value, attr.kind)
#                    mb_data_el.set_nested(attr, link)
#                else:
#                    mb_data_el.set_terminal_converted(attr, value)
#            if is_member_rpr:
#                result_data_el = mb_data_el
#            else:
#                coll_data_el.add_member(mb_data_el)
#        return result_data_el


class CsvData(object):
    def __init__(self, data=None):
        if data is None:
            data = {}
        self.fields = []
        self.data = []
        for attr_name, value in iteritems_(data):
            if not isinstance(value, CsvData):
                self.fields.append(attr_name)
                if len(self.data) == 0:
                    self.data.append([value])
                else:
                    for row in self.data:
                        row.append(value)
            else:
                self.expand(value)

    def expand(self, other):
        if len(self.data) == 0:
            self.data = other.data
        else:
            new_data = []
            for self_row, other_row in list(product(self.data, other.data)):
                new_data.append(self_row + other_row)
            self.data = new_data
        self.fields = self.fields + other.fields

    def append(self, other):
        if len(self.data) == 0:
            self.data = other.data
            self.fields = other.fields
        else:
            for row in other.data:
                self.data.append(row)

    def __len__(self):
        return len(self.data)


class CsvDataElementTreeVisitor(ResourceDataVisitor):
    def __init__(self, encoding):
        ResourceDataVisitor.__init__(self)
        self.__encoding = encoding
        self.__csv_data = None

    def visit_member(self, attribute_key, attribute, member_node, member_data,
                     is_link_node, parent_data, index=None):
        if is_link_node:
            new_field_name = self.__get_field_name(attribute_key.names[:-1],
                                                   attribute)
            mb_data = CsvData({new_field_name: member_node.get_url()})
        else:
            rpr_mb_data = OrderedDict()
            for attr, value in iteritems_(member_data):
                new_field_name = self.__get_field_name(attribute_key.names,
                                                       attr)
                rpr_mb_data[new_field_name] = self.__encode(value)
            mb_data = CsvData(rpr_mb_data)
        if not index is None:
            # Collection member. Store in parent data with index as key.
            parent_data[index] = mb_data
        elif len(attribute_key) == 0:
            # Top level - store as CSV data..
            self.__csv_data = mb_data
        else:
            # Nested member. Store in parent data with attribute as key.
            parent_data[attribute] = mb_data

    def visit_collection(self, attribute_key, attribute, collection_node,
                         collection_data, is_link_node, parent_data):
        if is_link_node:
            new_field_name = self.__get_field_name(attribute_key.names[:-1],
                                                   attribute)
            coll_data = CsvData({new_field_name:collection_node.get_url()})
        else:
            coll_data = CsvData()
            for item in sorted(collection_data.items()):
                mb_data = item[1]
                coll_data.append(mb_data)
        if len(attribute_key) == 0:
            self.__csv_data = coll_data
        else:
            parent_data[attribute] = coll_data

    @property
    def csv_data(self):
        return self.__csv_data

    def __get_field_name(self, attribute_names, attribute):
        if attribute.name != attribute.repr_name:
            field_name = attribute.repr_name
        else:
            field_name = '.'.join(attribute_names + (attribute.name,))
        return field_name # self.__encode(field_name)

    def __encode(self, item):
        if not PY3 and isinstance(item, text_type): # pragma: no cover
            item = item.encode(self.__encoding)
        return item


class CsvRepresentationGenerator(RepresentationGenerator):
    """
    A generator converting data elements into CSV representations.

    :note: ``None`` values in terminal attributes are represented as the empty
           string (this is the default behavior of the CSV writer from the
           standard library).
    :note: Nested member and collection resources are handled by adding
           more columns (member attributes) and rows (collection members)
           dynamically. By default, column names for nested member attributes
           are built as dot-concatenation of the corresponding attribute key.
    """
    def run(self, data_element):
        # We also emit None values to make sure every data row has the same
        # number of fields.
        trv = DataElementTreeTraverser(data_element, self._mapping,
                                       ignore_none_values=False)
        vst = CsvDataElementTreeVisitor(self.get_option('encoding'))
        trv.run(vst)
        csv_data = vst.csv_data
        if len(csv_data) > 0:
            wrt = writer(self._stream, dialect=self.get_option('dialect'))
            wrt.writerow(csv_data.fields)
            for row_data in csv_data.data:
                wrt.writerow(row_data)


class CsvResourceRepresenter(MappingResourceRepresenter):
    """
    Resource representer implementation for CSV.
    """
    content_type = CsvMime
    #: The CSV dialect to use for exporting CSV data.
    CSV_EXPORT_DIALECT = 'export'
    #: The CSV dialect to use for importing CSV data.
    CSV_IMPORT_DIALECT = 'import'

    @classmethod
    def make_mapping_registry(cls):
        return CsvMappingRegistry()

    if not PY3:
        # Under 2.x, the CSV writer and reader operates on byte streams, so
        # we have to override a few methods here.
        def to_bytes(self, obj, encoding=None):
            stream = NativeIO()
            self.to_stream(obj, stream)
            return stream.getvalue()

        def from_bytes(self, bytes_representation, resource=None):
            stream = NativeIO(bytes_representation)
            return self.from_stream(stream, resource=resource)

        def from_string(self, string_representation, resource=None):
            buf = bytes_(string_representation, encoding=self.encoding)
            stream = NativeIO(buf)
            return self.from_stream(stream, resource=resource)

    def _make_representation_parser(self, stream, resource_class, mapping):
        parser = CsvRepresentationParser(stream, resource_class, mapping)
        parser.set_option('dialect', self.CSV_IMPORT_DIALECT)
        parser.set_option('encoding', self.encoding)
        return parser

    def _make_representation_generator(self, stream, resource_class, mapping):
        generator = CsvRepresentationGenerator(stream, resource_class, mapping)
        generator.set_option('dialect', self.CSV_EXPORT_DIALECT)
        generator.set_option('encoding', self.encoding)
        return generator


class CsvMemberDataElement(SimpleMemberDataElement):
    converter_registry = CsvConverterRegistry


class CsvCollectionDataElement(SimpleCollectionDataElement):
    pass


class CsvLinkedDataElement(SimpleLinkedDataElement):
    pass


class CsvRepresenterConfiguration(RepresenterConfiguration):
    pass


class CsvMappingRegistry(SimpleMappingRegistry):
    """
    Registry for CSV mappings.
    """
    member_data_element_base_class = CsvMemberDataElement
    collection_data_element_base_class = CsvCollectionDataElement
    linked_data_element_base_class = CsvLinkedDataElement
    configuration_class = CsvRepresenterConfiguration



"""
RFC3339 time stamp generation support.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 3, 2013.

This is a fork of Henry Prechneur's rfc3339 module, stripped of test code,
doc strings and comments and adapted for Python 3.x compatibility. It will
be removed once the original package is also compatible with Python 3.x.

# Copyright (c) 2009, 2010, Henry Precheur <henry@precheur.org>
"""
import datetime
import time

__docformat__ = 'reStructuredText en'
__all__ = ['rfc3339']


def _timezone(utc_offset):
    hours = abs(utc_offset) // 3600
    minutes = abs(utc_offset) % 3600 // 60
    return '%c%02d:%02d' % ('-' if utc_offset < 0 else '+', hours, minutes)

def _timedelta_to_seconds(timedelta):
    return (timedelta.days * 86400 + timedelta.seconds +
            timedelta.microseconds // 1000)

def _utc_offset(date, use_system_timezone):
    if isinstance(date, datetime.datetime) and date.tzinfo is not None:
        return _timedelta_to_seconds(date.dst() or date.utcoffset())
    elif use_system_timezone:
        if date.year < 1970:
            # We use 1972 because 1970 doesn't have a leap day (feb 29)
            t = time.mktime(date.replace(year=1972).timetuple())
        else:
            t = time.mktime(date.timetuple())
        if time.localtime(t).tm_isdst: # pragma: no cover
            return -time.altzone
        else:
            return -time.timezone
    else:
        return 0

def _string(d, timezone):
    return ('%04d-%02d-%02dT%02d:%02d:%02d%s' %
            (d.year, d.month, d.day, d.hour, d.minute, d.second, timezone))

def rfc3339(date, utc=False, use_system_timezone=True):
    try:
        if use_system_timezone:
            date = datetime.datetime.fromtimestamp(date)
        else:
            date = datetime.datetime.utcfromtimestamp(date)
    except TypeError:
        pass

    if not isinstance(date, datetime.date):
        raise TypeError('Expected timestamp or date object. Got %r.' %
                        type(date))

    if not isinstance(date, datetime.datetime):
        date = datetime.datetime(*date.timetuple()[:3])
    utc_offset = _utc_offset(date, use_system_timezone)
    if utc:
        return _string(date + datetime.timedelta(seconds=utc_offset), 'Z')
    else:
        return _string(date, _timezone(utc_offset))
